#coding:utf-8

import datetime as dt
import threading
import time
import traceback

from xtquant import xtdata
from xtquant import xtbson as bson

_request_id = ''

def datetime_to_timetag(timelabel, format = ''):
    '''
    timelabel: str '20221231' '20221231235959'
    format: str '%Y%m%d' '%Y%m%d%H%M%S'
    return: int 1672502399000
    '''
    if not format:
        format = '%Y%m%d' if len(timelabel) == 8 else '%Y%m%d%H%M%S'
    return dt.datetime.strptime(timelabel, format).timestamp() * 1000

def timetag_to_datetime(timetag, format = ''):
    '''
    timetag: int 1672502399000
    format: str '%Y%m%d' '%Y%m%d%H%M%S'
    return: str '20221231' '20221231235959'
    '''
    if not format:
        format = '%Y%m%d' if timetag % 86400000 == 57600000 else '%Y%m%d%H%M%S'
    return dt.datetime.fromtimestamp(timetag / 1000).strftime(format)

def subscribe_quote(stock_code, period, dividend_type, count = 0, result_type = '', callback = None):
    return xtdata.subscribe_quote(stock_code, period, '', '', count, callback)

def subscribe_whole_quote(code_list, callback = None):
    return xtdata.subscribe_whole_quote(code_list, callback)

def unsubscribe_quote(subscribe_id):
    return xtdata.unsubscribe_quote(subscribe_id)

def get_market_data(
    fields = [], stock_code = [], start_time = '', end_time = ''
    , skip_paused = True, period = '', dividend_type = '', count = -1
):
    res = {}
    if period == 'tick':
        refixed = False
        if count == -2:
            refixed = True
            count = 1
        if 'quoter' not in fields:
            return xtdata.get_market_data_ori(
                    field_list=fields, stock_list=stock_code, period=period
                    , start_time=start_time, end_time=end_time, count=count
                    , dividend_type=dividend_type, fill_data=skip_paused
                )

        fields = []
        data = xtdata.get_market_data_ori(
            field_list=fields, stock_list=stock_code, period=period
            , start_time=start_time, end_time=end_time, count=count
            , dividend_type=dividend_type, fill_data=skip_paused
        )
        fields = ['quoter']

        import pandas as pd

        stime_fmt = '%Y%m%d' if period == '1d' else '%Y%m%d%H%M%S'
        for stock in data:
            pd_data = pd.DataFrame(data[stock])
            pd_data['stime'] = [timetag_to_datetime(t, stime_fmt) for t in pd_data['time']]
            pd_data.index = pd.to_datetime((pd_data['time'] + 28800000) * 1000000)
            ans = {}
            for j, timetag in enumerate(pd_data['time']):
                d_map = {}
                for key in pd_data:
                    d_map[key] = pd_data[key][j]
                ans[str(pd_data.index[j])] = {}
                ans[str(pd_data.index[j])]['quoter'] = d_map
            res[stock] = ans

        oriData = res
            # if not pd_data.empty:
            #     if count > 0:
            #         return list(pd_data.T.to_dict().values())
            #     return pd_data.iloc[-1].to_dict()
            # return {}
        if refixed:
            count = -2
    else:
        refixed = False
        if count == -2:
            refixed = True
            count = 1
        index, data = xtdata.get_market_data_ori(
            field_list=fields, stock_list=stock_code, period=period
            , start_time=start_time, end_time=end_time, count=count
            , dividend_type=dividend_type, fill_data=skip_paused
        )
        if refixed:
            end_time = ''
            count = -1
        for i, stock in enumerate(index[0]):
            ans = {}
            for j, timetag in enumerate(index[1]):
                d_map = {}
                for key in data:
                    d_map[key] = data[key][i][j]
                ans[timetag] = d_map
            res[stock] = ans
        oriData = res

    resultDict = {}
    for code in oriData:
        for timenode in oriData[code]:
            values = []
            for field in fields:
                values.append(oriData[code][timenode][field])
            key = code + timenode
            resultDict[key] = values

    if len(fields) == 1 and len(stock_code) <= 1 and (
            (start_time == '' and end_time == '') or start_time == end_time) and (count == -1 or count == -2):
        # if resultDict:
            # keys = list(resultDict.keys())
            # if resultDict[keys[-1]]:
            #     return resultDict[keys[-1]]
        for key in resultDict:
            return resultDict[key][0]
        return -1
    import numpy as np
    import pandas as pd
    if len(stock_code) <= 1 and start_time == '' and end_time == '' and (count == -1 or count == -2):
        for key in resultDict:
            result = pd.Series(resultDict[key], index=fields)
            return result
    if len(stock_code) > 1 and start_time == '' and end_time == '' and (count == -1 or count == -2):
        values = []
        for code in stock_code:
            if code in oriData:
                if not oriData[code]:
                    values.append([np.nan])
                for timenode in oriData[code]:
                    key = code + timenode
                    values.append(resultDict[key])
            else:
                values.append([np.nan])
        result = pd.DataFrame(values, index=stock_code, columns=fields)
        return result
    if len(stock_code) <= 1 and ((start_time != '' or end_time != '') or count >= 0):
        values = []
        times = []
        for code in oriData:
            for timenode in oriData[code]:
                key = code + timenode
                times.append(timenode)
                values.append(resultDict[key])
        result = pd.DataFrame(values, index=times, columns=fields)
        return result
    if len(stock_code) > 1 and ((start_time != '' or end_time != '') or count >= 0):
        values = {}
        for code in stock_code:
            times = []
            value = []
            if code in oriData:
                for timenode in oriData[code]:
                    key = code + timenode
                    times.append(timenode)
                    value.append(resultDict[key])
            values[code] = pd.DataFrame(value, index=times, columns=fields)
        try:
            result = pd.Panel(values)
            return result
        except:
            return oriData
    return

def get_market_data_ex(
    fields = [], stock_code = [], period = ''
    , start_time = '', end_time = '', count = -1
    , dividend_type = '', fill_data = True, subscribe = True
):
    res = xtdata.get_market_data_ex(
        field_list = fields, stock_list = stock_code, period = period
        , start_time = start_time, end_time = end_time, count = count
        , dividend_type = dividend_type, fill_data = fill_data
    )
    for stock in res:
        res[stock].index.name = "stime"
    return res

def get_full_tick(stock_code):
    return xtdata.get_full_tick(stock_code)

def get_divid_factors(stock_code, date = None):
    client = xtdata.get_client()
    if date:
        data = client.get_divid_factors(stock_code, date, date)
    else:
        data = client.get_divid_factors(stock_code, '19700101', '20380119')

    res = {}
    for value in data.values():
        res[value['time']] = list(value.values())[1:]
    return res

def download_history_data(stockcode, period, startTime, endTime):
    return xtdata.download_history_data(stockcode, period, startTime, endTime)

def get_raw_financial_data(field_list, stock_list, start_date, end_date, report_type = 'announce_time'):
    client = xtdata.get_client()
    data = client.get_financial_data(stock_list, field_list, start_date, end_date, report_type)

    import time
    res = {}
    for stock in data:
        stock_data = data[stock]
        res[stock] = {}

        for field in field_list:
            fs = field.split('.')
            table_data = stock_data[fs[0]]

            ans = {}
            for row_data in table_data:
                date = time.strftime('%Y%m%d', time.localtime(row_data[report_type] / 1000))
                if start_date <= date <= end_date:
                    ans[int(row_data[report_type])] = row_data[fs[1]]
            res[stock][field] = ans
    return res

#def download_financial_data(stock_list, table_list): #暂不提供
#    return xtdata.download_financial_data(stock_list, table_list)

def get_instrument_detail(stock_code, iscomplete = False):
    return xtdata.get_instrument_detail(stock_code, iscomplete)

#def get_instrument_type(stock_code): #暂不提供
#    return xtdata.get_instrument_type(stock_code)

def get_trading_dates(stock_code, start_date, end_date, count = -1, period = '1d'):
    if period != '1d':
        return []
    market = stock_code.split('.')[0]
    trade_dates = xtdata.get_trading_dates(market, start_date, end_date)
    if count == -1:
        return trade_dates
    if count > 0:
        return trade_dates[-count:]
    return []

def get_stock_list_in_sector(sector_name):
    return xtdata.get_stock_list_in_sector(sector_name)

def download_sector_data():
    return xtdata.download_sector_data()

download_sector_weight = download_sector_data #compat

def get_his_st_data(stock_code):
    return xtdata.get_his_st_data(stock_code)


def _passorder_impl(
    optype, ordertype, accountid, ordercode
    , prtype, modelprice, volume, ordertime, requestid
):
    data = {}

    data['optype'] = optype
    data['ordertype'] = ordertype
    data['accountid'] = accountid
    data['ordercode'] = ordercode
    data['prtype'] = prtype
    data['modelprice'] = modelprice
    data['volume'] = volume
    data['ordertime'] = ordertime

    client = xtdata.get_client()
    client.callFormula(requestid, 'passorder', bson.BSON.encode(data))
    return

def passorder(opType, orderType, accountid, orderCode, prType, modelprice, volume, C):
    return C.passorder(opType, orderType, accountid, orderCode, prType, modelprice, volume)

def get_trade_detail_data(accountid, accounttype, datatype, strategyname = ''):

    data = {}

    data['accountid'] = accountid
    data['accounttype'] = accounttype
    data['datatype'] = datatype
    data['strategyname'] = strategyname

    client = xtdata.get_client()
    result_bson = client.callFormula(_request_id, 'gettradedetail', bson.BSON.encode(data))
    result = bson.BSON.decode(result_bson)

    class DetailData(object):
        def __init__(self, _obj):
            if _obj:
                self.__dict__.update(_obj)

    out = []
    for item in result.get('result'):
        out.append(DetailData(item))
    return out

def register_external_resp_callback(reqid, callback):
    client = xtdata.get_client()

    status = [False, 0, 1, '']

    def on_callback(type, data, error):
        try:
            result = bson.BSON.decode(data)
            callback(type, result, error)
            return True
        except:
            status[0] = True
            status[3] = 'exception'
            return True

    client.register_external_resp_callback(reqid, on_callback)

def _set_auto_trade_callback_impl(enable, requestid):
    data = {}
    data['enable'] = enable

    client = xtdata.get_client()
    client.callFormula(requestid, 'setautotradecallback', bson.BSON.encode(data))
    return

def set_auto_trade_callback(C,enable):
    return C.set_auto_trade_callback(enable)

def set_account(accountid, requestid):
    data = {}
    data['accountid'] = accountid

    client = xtdata.get_client()
    client.callFormula(requestid, 'setaccount', bson.BSON.encode(data))
    return

def _get_callback_cache_impl(type, requestid):
    data = {}

    data['type'] = type

    client = xtdata.get_client()
    result_bson = client.callFormula(requestid, 'getcallbackcache', bson.BSON.encode(data))
    return bson.BSON.decode(result_bson)

def get_account_callback_cache(data, C):
    data = C.get_callback_cache("account").get('')
    return

def get_order_callback_cache(data, C):
    data = C.get_callback_cache("order")
    return

def get_deal_callback_cache(data, C):
    data = C.get_callback_cache("deal")
    return

def get_position_callback_cache(data, C):
    data = C.get_callback_cache("position")
    return

def get_ordererror_callback_cache(data, C):
    data = C.get_callback_cache("ordererror")
    return

def get_option_detail_data(stock_code):
    return xtdata.get_option_detail_data(stock_code)

def get_option_undl_data(undl_code_ref):
    return xtdata.get_option_undl_data(undl_code_ref)

def get_option_list(undl_code,dedate,opttype = "",isavailavle = False):
    return xtdata.get_option_list(undl_code, dedate, opttype, isavailavle)

def get_opt_iv(opt_code, requestid):
    data = {}
    data['code'] = opt_code

    client = xtdata.get_client()
    result_bson = client.callFormula(requestid, 'getoptiv', bson.BSON.encode(data))
    result = bson.BSON.decode(result_bson)

    out = result.get('result', 0)
    return out

def calc_bsm_price(optionType,strikePrice, targetPrice, riskFree, sigma, days, dividend, requestid):
    data = {}
    data['optiontype'] = optionType
    data['strikeprice'] = strikePrice
    data['targetprice'] = targetPrice
    data['riskfree'] = riskFree
    data['sigma'] = sigma
    data['days'] = days
    data['dividend'] = dividend

    client = xtdata.get_client()
    result_bson = client.callFormula(requestid, 'calcbsmprice', bson.BSON.encode(data))
    result = bson.BSON.decode(result_bson)

    out = result.get('result', 0)
    return out

def calc_bsm_iv(optionType, strikePrice, targetPrice, optionPrice, riskFree, days, dividend, requestid):
    data = {}
    data['optiontype'] = optionType
    data['strikeprice'] = strikePrice
    data['targetprice'] = targetPrice
    data['optionprice'] = optionPrice
    data['riskfree'] = riskFree
    data['days'] = days
    data['dividend'] = dividend

    client = xtdata.get_client()
    result_bson = client.callFormula(requestid, 'calcbsmiv', bson.BSON.encode(data))
    result = bson.BSON.decode(result_bson)

    out = result.get('result', 0)
    return out

def get_ipo_info(start_time, end_time):
    return xtdata.get_ipo_info(start_time, end_time)



# Copyright 2022-present MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Workaround for https://bugs.python.org/issue43923.
Ideally we would have done this with a single class, but
generic subclasses *must* take a parameter, and prior to Python 3.9
or in Python 3.7 and 3.8 with `from __future__ import annotations`,
you get the error: "TypeError: 'type' object is not subscriptable".
"""

import datetime
import abc
import enum
from typing import Tuple, Generic, Optional, Mapping, Any, TypeVar, Type, Dict, Iterable, Tuple, MutableMapping, Callable, Union


class TypeEncoder(abc.ABC, metaclass=abc.ABCMeta):
    @property
    @abc.abstractmethod
    def python_type(self) -> Any: ...
    @abc.abstractmethod
    def transform_python(self, value: Any) -> Any: ...

class TypeDecoder(abc.ABC, metaclass=abc.ABCMeta):
    @property
    @abc.abstractmethod
    def bson_type(self) -> Any: ...
    @abc.abstractmethod
    def transform_bson(self, value: Any) -> Any: ...

class TypeCodec(TypeEncoder, TypeDecoder, metaclass=abc.ABCMeta): ...

Codec = Union[TypeEncoder, TypeDecoder, TypeCodec]
Fallback = Callable[[Any], Any]

class TypeRegistry:
    _decoder_map: Dict[Any, Any]
    _encoder_map: Dict[Any, Any]
    _fallback_encoder: Optional[Fallback]

    def __init__(self, type_codecs: Optional[Iterable[Codec]] = ..., fallback_encoder: Optional[Fallback] = ...) -> None: ...
    def __eq__(self, other: Any) -> Any: ...


_DocumentType = TypeVar("_DocumentType", bound=Mapping[str, Any])

class DatetimeConversion(int, enum.Enum):
    DATETIME = ...
    DATETIME_CLAMP = ...
    DATETIME_MS = ...
    DATETIME_AUTO = ...

class CodecOptions(Tuple, Generic[_DocumentType]):
    document_class: Type[_DocumentType]
    tz_aware: bool
    uuid_representation: int
    unicode_decode_error_handler: Optional[str]
    tzinfo: Optional[datetime.tzinfo]
    type_registry: TypeRegistry
    datetime_conversion: Optional[int]

    def __new__(
        cls: Type[CodecOptions],
        document_class: Optional[Type[_DocumentType]] = ...,
        tz_aware: bool = ...,
        uuid_representation: Optional[int] = ...,
        unicode_decode_error_handler: Optional[str] = ...,
        tzinfo: Optional[datetime.tzinfo] = ...,
        type_registry: Optional[TypeRegistry] = ...,
        datetime_conversion: Optional[int] = ...,
    ) -> CodecOptions[_DocumentType]: ...

    # CodecOptions API
    def with_options(self, **kwargs: Any) -> CodecOptions[_DocumentType]: ...

    def _arguments_repr(self) -> str: ...

    def _options_dict(self) -> Dict[Any, Any]: ...

    # NamedTuple API
    @classmethod
    def _make(cls, obj: Iterable) -> CodecOptions[_DocumentType]: ...

    def _asdict(self) -> Dict[str, Any]: ...

    def _replace(self, **kwargs: Any) -> CodecOptions[_DocumentType]: ...

    _source: str
    _fields: Tuple[str]


DEFAULT_CODEC_OPTIONS: CodecOptions[MutableMapping[str, Any]]
_RAW_BSON_DOCUMENT_MARKER: int

def _raw_document_class(document_class: Any) -> bool: ...

def _parse_codec_options(options: Any) -> CodecOptions: ...



#coding:utf-8

import os, sys
import ctypes as ct

import json
from . import xtbson as bson


def read_from_bson_buffer(buffer):
    result = []

    pos = 0
    while 1:
        if pos + 4 < len(buffer):
            dlen_buf = buffer[pos : pos + 4]
        else:
            break

        dlen = ct.cast(dlen_buf, ct.POINTER(ct.c_int32))[0]
        if dlen >= 5:
            try:
                data_buf = buffer[pos : pos + dlen]
                pos += dlen

                result.append(bson.decode(data_buf))
            except Exception as e:
                pass
        else:
            break

    return result


def write_to_bson_buffer(data_list):
    buffer = b''

    for data in data_list:
        buffer += bson.encode(data)

    return buffer


def read_from_feather_file(file):
    import feather as fe
    meta = {}
    return meta, fe.read_dataframe(file)


def write_to_feather_file(data, file, meta = None):
    import feather as fe
    if not meta:
        meta = {}
    return




XT_GE_BANK_LOAN= 10051
XT_GE_BJ= 20002
XT_GE_BOND_DISTRIBUTION= 100200
XT_GE_DF_ARBITAGE_FTOPTION= 100056
XT_GE_DF_FTOPTION= 100052
XT_GE_EXTRA_300_ETF= 100101
XT_GE_EXTRA_50_ETF= 100100
XT_GE_EXTRA_ALLOW_PLEDGE= 100105
XT_GE_EXTRA_BLOCK_TRADING= 100102
XT_GE_EXTRA_CB_ETF= 100026
XT_GE_EXTRA_CB_LOF= 100027
XT_GE_EXTRA_CLOSED_ENDED_FUNDS= 100014
XT_GE_EXTRA_DIRECTIONAL_CONVERTIBALE_BOND= 100104
XT_GE_EXTRA_ETF= 100013
XT_GE_EXTRA_FICC= 100090
XT_GE_EXTRA_FI_ETF= 100034
XT_GE_EXTRA_FUND= 100004
XT_GE_EXTRA_GF= 100036
XT_GE_EXTRA_GLR= 100020
XT_GE_EXTRA_GOLD= 100022
XT_GE_EXTRA_INTRA_DAY= 100030
XT_GE_EXTRA_LOAN= 100016
XT_GE_EXTRA_LOAN_ETF= 100038
XT_GE_EXTRA_LOF= 100037
XT_GE_EXTRA_MAIN_BOARD= 100029
XT_GE_EXTRA_MARKET_CF= 100006
XT_GE_EXTRA_MARKET_FU= 100007
XT_GE_EXTRA_MARKET_ST= 100008
XT_GE_EXTRA_MONETARY_FUND= 100024
XT_GE_EXTRA_NOT_CLOSING_AUCTION_MATCH= 100070
XT_GE_EXTRA_POLICY_JRZ= 100019
XT_GE_EXTRA_PUBLIC_INFRASTRUCTURE_FUND= 100103
XT_GE_EXTRA_RATE_BOND= 100080
XT_GE_EXTRA_REPURCHASE_DAY_1= 100040
XT_GE_EXTRA_REPURCHASE_DAY_14= 100045
XT_GE_EXTRA_REPURCHASE_DAY_2= 100041
XT_GE_EXTRA_REPURCHASE_DAY_28= 100046
XT_GE_EXTRA_REPURCHASE_DAY_28_UPPER= 100047
XT_GE_EXTRA_REPURCHASE_DAY_3= 100042
XT_GE_EXTRA_REPURCHASE_DAY_4= 100043
XT_GE_EXTRA_REPURCHASE_DAY_7= 100044
XT_GE_EXTRA_REPURCHASE_IMPAWN= 100021
XT_GE_EXTRA_RTMF= 100023
XT_GE_EXTRA_SH_CORPORATE_LOAN= 100106
XT_GE_EXTRA_SH_DISTRIBUTION= 100031
XT_GE_EXTRA_SH_IPO= 100025
XT_GE_EXTRA_SH_LOAN= 100039
XT_GE_EXTRA_STANDARD_BOND= 100018
XT_GE_EXTRA_STOCK= 100003
XT_GE_EXTRA_STOCK_A= 100001
XT_GE_EXTRA_STOCK_B= 100002
XT_GE_EXTRA_STOCK_EX= 100012
XT_GE_EXTRA_STOCK_INDEX= 100005
XT_GE_EXTRA_STOCK_TRANABLE= 100028
XT_GE_EXTRA_ST_FIX= 100035
XT_GE_EXTRA_SZ_CGL= 100009
XT_GE_EXTRA_SZ_CORPORATE_LOAN= 100107
XT_GE_EXTRA_SZ_GL= 100010
XT_GE_EXTRA_SZ_GLR= 100017
XT_GE_EXTRA_SZ_LOAN= 100011
XT_GE_EXTRA_WARRANT= 100015
XT_GE_EXTRA_XGED= 100032
XT_GE_IF_ARBITAGE_FTOPTION= 100057
XT_GE_IF_FTOPTION= 100053
XT_GE_INE_FTOPTION= 100058
XT_GE_MARKET_DS= 5
XT_GE_MARKET_NEW3BOARD= 9
XT_GE_MARKET_NEW3BOARD_DELISTED= 20000
XT_GE_MARKET_OF= 7
XT_GE_MARKET_OP= 8
XT_GE_MARKET_SH= 1
XT_GE_MARKET_SQ= 4
XT_GE_MARKET_SZ= 2
XT_GE_MARKET_ZJ= 3
XT_GE_MARKET_ZS= 6
XT_GE_NEW3BOARD_PREFERED_SHARES_TRANSFER= 20001
XT_GE_SF_ARBITAGE_FTOPTION= 100054
XT_GE_SF_FIXED_INCOME_ETF= 155
XT_GE_SF_FTOPTION= 100050
XT_GE_SHARES_ALLOTMEN= 100033
XT_GE_SH_300_ETF= 351
XT_GE_SH_50_ETF= 350
XT_GE_SH_A= 101
XT_GE_SH_ABS= 113
XT_GE_SH_ABS_TRANSFER= 300
XT_GE_SH_ALLOW_PLEDGE_BOND= 359
XT_GE_SH_ASS= 186
XT_GE_SH_ASSET_BACKED_SECURITIES= 242
XT_GE_SH_ASSET_SECURITIZATION= 132
XT_GE_SH_B= 102
XT_GE_SH_BONDS_FUNDS= 166
XT_GE_SH_BOND_OFFER_REPURCHASE= 150
XT_GE_SH_BOND_PROTOCOL_REPURCHASE= 215
XT_GE_SH_BOND_RESALE= 214
XT_GE_SH_BOND_TRIPARTITE_REPURCHASE= 216
XT_GE_SH_CDR= 196
XT_GE_SH_CDR_ALLOTMEN= 199
XT_GE_SH_CDR_DISTRIBUTION= 197
XT_GE_SH_CDR_SUBSCRIBE= 198
XT_GE_SH_CENTRAL_GOVERNMENT_LOAN= 130
XT_GE_SH_CLOSED_ENDED_FUNDS= 140
XT_GE_SH_CONVERTIBALE_BOND= 135
XT_GE_SH_CONVERTIBALE_BOND_RESALE= 239
XT_GE_SH_CONVERTIBLE_BOND_STOCK= 213
XT_GE_SH_CONVERTIBLE_BOUND_DISTRIBUTION= 164
XT_GE_SH_CORPORATE_BOND= 115
XT_GE_SH_CORPORATE_BOND_REPURCHASE= 145
XT_GE_SH_CORPORATE_LOAN_PLEDGE= 114
XT_GE_SH_COUPON_GOVERNMENT_LOAN_PLEDGE= 238
XT_GE_SH_CPB= 138
XT_GE_SH_CPB_LOAN= 176
XT_GE_SH_CPB_SWAP= 360
XT_GE_SH_CREDIT_ASSET_SUPPORTED= 133
XT_GE_SH_CROSS_BORDER_ETF= 152
XT_GE_SH_CROSS_BORDER_LOF= 153
XT_GE_SH_CR_ETF= 244
XT_GE_SH_CR_ETF_CROSS_MARKET= 250
XT_GE_SH_CR_FUND_ETF_GOLD= 274
XT_GE_SH_DESIGNATED_TRANSACTION= 234
XT_GE_SH_DIRECTIONAL_CONVERTIBALE_BOND= 303
XT_GE_SH_ENTERPRISE_BOND= 134
XT_GE_SH_ENTERPRISE_LOAN_PLEDGE= 240
XT_GE_SH_ENTERPROSE_SUPPORT_AUCTION_BOND= 299
XT_GE_SH_ENTERPROSE_SUPPORT_BOND= 191
XT_GE_SH_EPB_TRANSFER= 175
XT_GE_SH_ETF= 105
XT_GE_SH_EXCHANGEABLE_BOND_DISTRIBUTION= 233
XT_GE_SH_EXCHANGEABLE_LOAN= 108
XT_GE_SH_EXCHANGEABLE_LOAN_PLEDGE= 109
XT_GE_SH_EXCHANGEABLE_LOAN_SWAP= 110
XT_GE_SH_FINANCIAL_BONDS= 128
XT_GE_SH_FUND= 103
XT_GE_SH_FUND_ETF= 245
XT_GE_SH_FUND_ETF_CROSS_BORDER= 255
XT_GE_SH_FUND_ETF_CROSS_MARKET= 251
XT_GE_SH_FUND_ETF_CR_LOAN= 264
XT_GE_SH_FUND_IPO= 161
XT_GE_SH_FUND_RAISING= 223
XT_GE_SH_FUND_SUBSCRIPTION= 227
XT_GE_SH_FUND_SUBSCRIPTION_DISTRIBUTION= 228
XT_GE_SH_GF= 170
XT_GE_SH_GOLD= 156
XT_GE_SH_GOVERNMENT_BANK_FINANCE_LOAN_DISTRIBUTE_SALE= 361
XT_GE_SH_GOVERNMENT_LOAN= 137
XT_GE_SH_GOVERNMENT_LOAN_DISCOUNT= 129
XT_GE_SH_GOVERNMENT_LOAN_DISCOUNT_REPURCHASE= 144
XT_GE_SH_GOVERNMENT_LOAN_DISTRIBUTE_SALE= 146
XT_GE_SH_GOVERNMENT_LOAN_FUTURES= 217
XT_GE_SH_GOVERNMENT_LOAN_INTEREST_BEARING= 127
XT_GE_SH_GOVERNMENT_LOAN_PLEDGE= 237
XT_GE_SH_GOVERNMENT_LOAN_REPURCHASE= 126
XT_GE_SH_GOVERNMENT_LOAN_REPURCHASE_BUYOUT= 124
XT_GE_SH_GOVERNMENT_LOAN_REPURCHASE_ENTERPRISE= 123
XT_GE_SH_GOVERNMENT_LOAN_REPURCHASE_IMPAWN= 125
XT_GE_SH_GOVERNMENT_LOAN_REPURCHASE_TRUSTEE_SHIP= 122
XT_GE_SH_GSF= 185
XT_GE_SH_INDEX= 104
XT_GE_SH_INNOVATE_ALLOTMEN= 195
XT_GE_SH_INNOVATE_DISTRIBUTION= 193
XT_GE_SH_INNOVATION_CLOSED_ENDED_FUNDS= 154
XT_GE_SH_INTEREST_RATE_GOVERNMENT_LOAN= 229
XT_GE_SH_LOAN_CBB_SCB_NEW= 187
XT_GE_SH_LOAN_CR_ETF= 200
XT_GE_SH_LOAN_ETF= 174
XT_GE_SH_LOAN_IPO= 160
XT_GE_SH_LOAN_ISSUANCE_DISTRIBUTE_SALE= 232
XT_GE_SH_LOAN_REPURCHASE_DAY_1= 177
XT_GE_SH_LOAN_REPURCHASE_DAY_14= 182
XT_GE_SH_LOAN_REPURCHASE_DAY_2= 178
XT_GE_SH_LOAN_REPURCHASE_DAY_28= 183
XT_GE_SH_LOAN_REPURCHASE_DAY_28_UPPER= 184
XT_GE_SH_LOAN_REPURCHASE_DAY_3= 179
XT_GE_SH_LOAN_REPURCHASE_DAY_4= 180
XT_GE_SH_LOAN_REPURCHASE_DAY_7= 181
XT_GE_SH_LOCAL_GOVERNMENT_LOAN= 136
XT_GE_SH_LOCAL_GOVERNMENT_LOAN_DISTRIBUTE_SALE= 148
XT_GE_SH_LOCAL_GOVERNMENT_LOAN_ONLINE_DISTRIBUTE_SALE= 362
XT_GE_SH_LOCAL_GOVERNMENT_LOAN_REPURCHASE= 147
XT_GE_SH_LOF= 169
XT_GE_SH_LOW_CORPORATE_BOND= 241
XT_GE_SH_MONETARY_FUND_SUBSCRIPTION= 275
XT_GE_SH_MS_PRIVATE_PLACEMENT_BOND= 151
XT_GE_SH_NEW_SHARES_DISTRIBUTION= 162
XT_GE_SH_NON_FUND_ETF= 248
XT_GE_SH_NON_FUND_ETF_CROSS_BORDER= 258
XT_GE_SH_NON_FUND_ETF_CROSS_MARKET= 254
XT_GE_SH_NON_FUND_ETF_CR_LOAN= 267
XT_GE_SH_NON_FUND_ETF_LOAN= 262
XT_GE_SH_NON_FUND_ETF_MONETARY_FUND= 271
XT_GE_SH_NON_PUBLIC_CONVERTIBLE_BOND_STOCK= 235
XT_GE_SH_NON_PUBLIC_CONVERTIBLE_CORPORATE_LOAN= 354
XT_GE_SH_NON_PUBLIC_CORPORATE_LOAN= 298
XT_GE_SH_NON_PUBLIC_PREFERED_SHARES_TRANSFER= 117
XT_GE_SH_OLD_GOVERNMENT_LOAN= 143
XT_GE_SH_ONLINE_VOTING= 225
XT_GE_SH_ONLINE_VOTING_B= 226
XT_GE_SH_ONLINE_VOTING_PASSWORD_SERVICE_B= 236
XT_GE_SH_OPEN_END_FUND= 189
XT_GE_SH_OPEN_END_FUND_CONVERSION= 222
XT_GE_SH_OPEN_END_FUND_CROSS_MARKET= 220
XT_GE_SH_OPEN_END_FUND_DIVIDEND= 221
XT_GE_SH_OPEN_END_FUND_SUBSCRIPTION= 219
XT_GE_SH_OPEN_GF_LOF= 218
XT_GE_SH_OTHER_MF= 190
XT_GE_SH_PLACING= 363
XT_GE_SH_PLACING_FIRST_DISTRIBUTION= 163
XT_GE_SH_PLEDGE= 142
XT_GE_SH_POLICY_JRZ= 141
XT_GE_SH_PRICE_GOVERNMENT_LOAN= 230
XT_GE_SH_PRIVATELY_LOAN_TRANSFER= 111
XT_GE_SH_PR_CB_ETF= 173
XT_GE_SH_PUBLIC_CORPORATE_TRADE_LOAN= 352
XT_GE_SH_PUBLIC_INFRASTRUCTURE_FUND= 302
XT_GE_SH_PUBLIC_LOAN_DISTRIBUTE_SALE= 231
XT_GE_SH_PUBLIC_LOAN_PLEDGE= 212
XT_GE_SH_PUBLIC_PREFERED_SHARES= 116
XT_GE_SH_PUBLIC_PREFERED_SHARES_PLACEMENTS= 119
XT_GE_SH_PUBLIC_PREFERED_SHARES_SUBSCRIBE= 118
XT_GE_SH_PUBLIC_PREFERED_SHARES_SUBSCRIBE_DISTRIBUTION= 121
XT_GE_SH_PUBLIC_PREFERED_SHARES_SUBSCRIBE_PRICE= 120
XT_GE_SH_QUOTATION_REPURCHASE_DAY_1= 201
XT_GE_SH_QUOTATION_REPURCHASE_DAY_119= 208
XT_GE_SH_QUOTATION_REPURCHASE_DAY_14= 203
XT_GE_SH_QUOTATION_REPURCHASE_DAY_154= 209
XT_GE_SH_QUOTATION_REPURCHASE_DAY_182= 210
XT_GE_SH_QUOTATION_REPURCHASE_DAY_21= 292
XT_GE_SH_QUOTATION_REPURCHASE_DAY_210= 294
XT_GE_SH_QUOTATION_REPURCHASE_DAY_245= 295
XT_GE_SH_QUOTATION_REPURCHASE_DAY_273= 211
XT_GE_SH_QUOTATION_REPURCHASE_DAY_28= 204
XT_GE_SH_QUOTATION_REPURCHASE_DAY_301= 296
XT_GE_SH_QUOTATION_REPURCHASE_DAY_35= 293
XT_GE_SH_QUOTATION_REPURCHASE_DAY_357= 297
XT_GE_SH_QUOTATION_REPURCHASE_DAY_42= 205
XT_GE_SH_QUOTATION_REPURCHASE_DAY_63= 206
XT_GE_SH_QUOTATION_REPURCHASE_DAY_7= 202
XT_GE_SH_QUOTATION_REPURCHASE_DAY_91= 207
XT_GE_SH_RTMF= 157
XT_GE_SH_SEO= 172
XT_GE_SH_SEPERATION_BOND= 131
XT_GE_SH_SEPERATION_BOND_REPURCHASE= 149
XT_GE_SH_SHARES_ALLOTMEN= 167
XT_GE_SH_SHARES_CONVERTIBLE_BOND= 168
XT_GE_SH_SHORTTERM_CORPORATE_LOAN_TRANSFER= 112
XT_GE_SH_STANDARD_BOND= 139
XT_GE_SH_STOCK_IPO= 159
XT_GE_SH_SUBSCRIBE= 107
XT_GE_SH_SUBSCRIPTION_ETF= 246
XT_GE_SH_SUBSCRIPTION_ETF_CROSS_BORDER= 256
XT_GE_SH_SUBSCRIPTION_ETF_CROSS_MARKET= 252
XT_GE_SH_SUBSCRIPTION_ETF_CR_LOAN= 265
XT_GE_SH_SUBSCRIPTION_ETF_GOLD= 272
XT_GE_SH_SUBSCRIPTION_ETF_LOAN= 260
XT_GE_SH_SUBSCRIPTION_ETF_MONETARY_FUND= 269
XT_GE_SH_SUBSCRIPTION_FUND_ETF_GOLD= 273
XT_GE_SH_SUBSCRIPTION_PRICE= 165
XT_GE_SH_SUBSCRIPTION_REPAYMENT_ETF= 247
XT_GE_SH_SUBSCRIPTION_REPAYMENT_ETF_CROSS_BORDER= 257
XT_GE_SH_SUBSCRIPTION_REPAYMENT_ETF_CROSS_MARKET= 253
XT_GE_SH_SUBSCRIPTION_REPAYMENT_ETF_CR_LOAN= 266
XT_GE_SH_SUBSCRIPTION_REPAYMENT_ETF_LOAN= 261
XT_GE_SH_SUBSCRIPTION_REPAYMENT_ETF_MONETARY_FUND= 270
XT_GE_SH_SUBSCRIPTION_TECH_BOARD= 278
XT_GE_SH_TECH_BOARD= 277
XT_GE_SH_TECH_BOARD_CDR= 288
XT_GE_SH_TECH_BOARD_CDR_ALLOTMEN= 291
XT_GE_SH_TECH_BOARD_CDR_DISTRIBUTION= 289
XT_GE_SH_TECH_BOARD_CDR_SUBSCRIBE= 290
XT_GE_SH_TECH_BOARD_CONVERTIBLE_BOND= 355
XT_GE_SH_TECH_BOARD_CONVERTIBLE_BOND_IPO= 356
XT_GE_SH_TECH_BOARD_CONVERTIBLE_BOND_IPO_DISTRIBUTION= 357
XT_GE_SH_TECH_BOARD_ETF= 353
XT_GE_SH_TECH_BOARD_ETF_ETFCODE= 364
XT_GE_SH_TECH_BOARD_SHARES_CONVERTIBLE_BOND= 358
XT_GE_SH_TENDER_OFFER= 224
XT_GE_SH_TMF= 158
XT_GE_SH_TMFR= 301
XT_GE_SH_TRANSACTION_ETF= 243
XT_GE_SH_TRANSACTION_ETF_CROSS_MARKET= 249
XT_GE_SH_TRANSACTION_ETF_CR_LOAN= 263
XT_GE_SH_TRANSACTION_ETF_LOAN= 259
XT_GE_SH_TRANSACTION_ETF_MONETARY_FUND= 268
XT_GE_SH_WARRANT= 106
XT_GE_SH_WARRANT_CREATION_CANCEL= 276
XT_GE_SH_XGED= 171
XT_GE_SZ_300_ETF= 10071
XT_GE_SZ_A= 10001
XT_GE_SZ_ASS= 10027
XT_GE_SZ_B= 10002
XT_GE_SZ_CB= 10013
XT_GE_SZ_CBB= 10019
XT_GE_SZ_CB_ETF= 10029
XT_GE_SZ_CB_LOF= 10030
XT_GE_SZ_CDR_ALLOTMEN= 10056
XT_GE_SZ_CEF= 10021
XT_GE_SZ_CYB_CCB= 10033
XT_GE_SZ_DIRECTIONAL_CONVERTIBALE_BOND= 10077
XT_GE_SZ_EB= 10014
XT_GE_SZ_ENTERPROSE_SUPPORT_BOND= 10055
XT_GE_SZ_ETF= 10008
XT_GE_SZ_ETF_CROSS_MARKET= 10079
XT_GE_SZ_EXCHANGEABLE_LOAN= 10078
XT_GE_SZ_FUND= 10003
XT_GE_SZ_GEM_BORAD= 10006
XT_GE_SZ_GEM_BORAD_DR= 10073
XT_GE_SZ_GF= 10023
XT_GE_SZ_GLD= 10012
XT_GE_SZ_GLIB= 10011
XT_GE_SZ_GLR= 10010
XT_GE_SZ_GLRA= 10035
XT_GE_SZ_GOLD= 10036
XT_GE_SZ_GOV_ALLOW= 10052
XT_GE_SZ_GSF= 10028
XT_GE_SZ_ICEF= 10031
XT_GE_SZ_INDEX= 10007
XT_GE_SZ_INNOVATE_KZZ= 10053
XT_GE_SZ_LGL= 10018
XT_GE_SZ_LOAN_DIRECTIONAL= 10069
XT_GE_SZ_LOAN_ETF= 10041
XT_GE_SZ_LOAN_IPO= 10054
XT_GE_SZ_LOAN_ISSUANCE_DISTRIBUTE_SALE= 10067
XT_GE_SZ_LOAN_REITS= 10068
XT_GE_SZ_LOAN_REPURCHASE_DAY_1= 10042
XT_GE_SZ_LOAN_REPURCHASE_DAY_14= 10047
XT_GE_SZ_LOAN_REPURCHASE_DAY_2= 10043
XT_GE_SZ_LOAN_REPURCHASE_DAY_28= 10048
XT_GE_SZ_LOAN_REPURCHASE_DAY_28_UPPER= 10049
XT_GE_SZ_LOAN_REPURCHASE_DAY_3= 10044
XT_GE_SZ_LOAN_REPURCHASE_DAY_4= 10045
XT_GE_SZ_LOAN_REPURCHASE_DAY_7= 10046
XT_GE_SZ_LOF= 10022
XT_GE_SZ_MAIN_BOARD= 10004
XT_GE_SZ_MAIN_SME_BORAD_DR= 10074
XT_GE_SZ_MSP_PB= 10016
XT_GE_SZ_NON_PUBLIC_PREFERED_SHARES_TRANSFER= 10066
XT_GE_SZ_OLDSHARES_PREFERRED_CONVERTIBLE_BOND= 10072
XT_GE_SZ_POB= 10050
XT_GE_SZ_PUBLIC_INFRASTRUCTURE_FUND= 10076
XT_GE_SZ_RTMF= 10037
XT_GE_SZ_SA= 10040
XT_GE_SZ_SB= 10015
XT_GE_SZ_SCB_PB= 10024
XT_GE_SZ_SC_SB= 10025
XT_GE_SZ_SEO= 10039
XT_GE_SZ_SFMP= 10017
XT_GE_SZ_SHARES_CONVERTIBLE_BOND= 10075
XT_GE_SZ_SME_BOARD= 10005
XT_GE_SZ_SPB= 10026
XT_GE_SZ_STANDAR_B= 10020
XT_GE_SZ_WARRANT= 10009
XT_GE_SZ_XGED= 10038
XT_GE_SZ_ZB_CCB= 10032
XT_GE_SZ_ZXB_CCB= 10034
XT_GE_ZF_ARBITAGE_FTOPTION= 100055
XT_GE_ZF_FTOPTION= 100051
XT_PLEDGE_REPURCHASE_SH_DAY_1= 279
XT_PLEDGE_REPURCHASE_SH_DAY_14= 281
XT_PLEDGE_REPURCHASE_SH_DAY_180= 285
XT_PLEDGE_REPURCHASE_SH_DAY_21= 282
XT_PLEDGE_REPURCHASE_SH_DAY_270= 286
XT_PLEDGE_REPURCHASE_SH_DAY_30= 283
XT_PLEDGE_REPURCHASE_SH_DAY_365= 287
XT_PLEDGE_REPURCHASE_SH_DAY_7= 280
XT_PLEDGE_REPURCHASE_SH_DAY_90= 284
XT_PLEDGE_REPURCHASE_SZ_DAY_1= 10057
XT_PLEDGE_REPURCHASE_SZ_DAY_14= 10059
XT_PLEDGE_REPURCHASE_SZ_DAY_180= 10063
XT_PLEDGE_REPURCHASE_SZ_DAY_21= 10060
XT_PLEDGE_REPURCHASE_SZ_DAY_270= 10064
XT_PLEDGE_REPURCHASE_SZ_DAY_30= 10061
XT_PLEDGE_REPURCHASE_SZ_DAY_365= 10065
XT_PLEDGE_REPURCHASE_SZ_DAY_7= 10058
XT_PLEDGE_REPURCHASE_SZ_DAY_90= 10062



#coding:utf-8

import os, sys
#import traceback
import json


### config

__curdir = os.path.dirname(os.path.abspath(__file__))

__rpc_config_file = os.path.join(__curdir, 'xtdata.ini')

from .IPythonApiClient import rpc_init
__rpc_init_status = rpc_init(__rpc_config_file)

if __rpc_init_status < 0:
    raise Exception(f'rpc init failed, error_code:{__rpc_init_status}, config:{__rpc_config_file}')


### function

def try_create_client():
    from .IPythonApiClient import IPythonApiClient as RPCClient

    cl = RPCClient()
    cl.init()

    ec = cl.load_config(__rpc_config_file, 'client_xtdata')
    if ec < 0:
        raise f'load config failed, file:{__rpc_config_file}'
    return cl


def try_create_connection(addr):
    '''
    addr: 'localhost:58610'
    '''
    ip, port = addr.split(':')
    if not ip:
        ip = 'localhost'
    if not port:
        raise Exception('invalid port')

    cl = try_create_client()
    cl.set_config_addr(addr)

    ec, msg = cl.connect()
    if ec < 0:
        raise Exception((ec, msg))
    return cl


def create_connection(addr):
    try:
        return try_create_connection(addr)
    except Exception as e:
        return None


def scan_available_server():
    try:
        result = []

        config_dir = os.path.abspath(os.path.join(os.environ['USERPROFILE'], '.xtquant'))

        for f in os.scandir(config_dir):
            full_path = f.path

            is_running = False
            try:
                os.remove(os.path.join(full_path, 'running_status'))
            except PermissionError:
                is_running = True
            except Exception as e:
                pass

            if not is_running:
                continue

            try:
                config = json.load(open(os.path.join(full_path, 'xtdata.cfg'), 'r', encoding = 'utf-8'))

                ip = config.get('ip', 'localhost')
                port = config.get('port', None)
                if not port:
                    raise Exception(f'invalid port: {port}')

                addr = f'{ip}:{port}'
                result.append(addr)
            except Exception as e:
                continue

        result.sort()
        return result

    except Exception as e:
        return []


def connect_any(addr_list):
    '''
    addr_list: [ addr, ... ]
        addr: 'localhost:58610'
    '''
    for addr in addr_list:
        try:
            cl = create_connection(addr)
            if cl:
                return cl
        except Exception as e:
            continue

    return None







#coding:utf-8

import os, sys
import datetime as dt
import traceback
import json

from . import xtbson as bson

### connection

__client = None
__client_last_spec = ('', None)


def connect(ip = '', port = None, remember_if_success = True):
    global __client

    if __client:
        if __client.is_connected():
            return __client

        __client.shutdown()
        __client = None

    from . import xtconn

    if not ip:
        ip = 'localhost'

    if port:
        server_list = [f'{ip}:{port}']
        __client = xtconn.connect_any(server_list)
    else:
        server_list = xtconn.scan_available_server()

        default_addr = 'localhost:58610'
        if not default_addr in server_list:
            server_list.append(default_addr)

        __client = xtconn.connect_any(server_list)

    if not __client or not __client.is_connected():
        raise Exception("无法连接xtquant服务，请检查QMT-投研版或QMT-极简版是否开启")

    if remember_if_success:
        global __client_last_spec
        __client_last_spec = (ip, port)

    return __client


def reconnect(ip = '', port = None, remember_if_success = True):
    global __client

    if __client:
        __client.shutdown()
        __client = None

    return connect(ip, port, remember_if_success)


def get_client():
    global __client

    if not __client or not __client.is_connected():
        global __client_last_spec

        ip, port = __client_last_spec
        __client = connect(ip, port, False)

    return __client


### utils
def try_except(func):
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except Exception:
            exc_type, exc_instance, exc_traceback = sys.exc_info()
            formatted_traceback = ''.join(traceback.format_tb(exc_traceback))
            message = '\n{0} raise {1}:{2}'.format(
                formatted_traceback,
                exc_type.__name__,
                exc_instance
            )
            # raise exc_type(message)
            print(message)
            return None

    return wrapper

def __bsoncall_common(interface, func, param):
    return bson.BSON.decode(interface(func, bson.BSON.encode(param)))

def create_view(viewID, view_type, title, group_id):
    client = get_client()
    return client.createView(viewID, view_type, title, group_id)

#def reset_view(viewID):
#    return

def close_view(viewID):
    client = get_client()
    return client.closeView(viewID)

#def set_view_index(viewID, datas):
#    '''
#    设置模型指标属性
#    index: { "output1": { "datatype": se::OutputDataType } }
#    '''
#    client = get_client()
#    return client.setViewIndex(viewID, datas)

def push_view_data(viewID, datas):
    '''
    推送模型结果数据
    datas: { "timetags: [t1, t2, ...], "outputs": { "output1": [value1, value2, ...], ... }, "overwrite": "full/increase" }
    '''
    client = get_client()
    bresult = client.pushViewData(viewID, 'index', bson.BSON.encode(datas))
    return bson.BSON.decode(bresult)

def switch_graph_view(stock_code = None, period = None, dividendtype = None, graphtype = None):
    cl = get_client()

    result = __bsoncall_common(
        cl.commonControl, 'switchgraphview'
        , {
            "stockcode": stock_code
            , "period": period
            , "dividendtype": dividendtype
            , "graphtype": graphtype
        }
    )


def push_xtview_data(data_type, time, datas):
    cl = get_client()
    timeData = 0
    types = []
    numericDatas = []
    stringDatas = []
    if type(time) == int:
        name_list = list(datas.keys())
        value_list = []
        for name in name_list:
            value_list.append([datas[name]])
        timeData = [time]
    if type(time) == list:
        time_list = time
        name_list = list(datas.keys())
        value_list = list(datas.values())
        timeData = time_list

    for value in value_list:
        if isinstance(value[0], str):
            stringDatas.append(value)
            types.append(3)
        else:
            numericDatas.append(value)
            types.append(0)

    result = __bsoncall_common(
        cl.custom_data_control, 'pushxtviewdata'
        , {
            'dataType': data_type
            ,'timetags': timeData
            , 'names' : name_list
            , 'types' : types
            , 'numericDatas' : numericDatas
            , 'stringDatas' : stringDatas
        }
    )
    return




#coding:utf-8

from . import contextinfo
from . import functions

import os
import uuid
import datetime as dt

from xtquant import xtdata
from xtquant import xtbson as bson
from xtquant import xtdata_config

class StrategyLoader:
    def __init__(this):
        this.C = None
        this.main_quote_subid = 0
        return

    def init(this):
        C = this.C

        C.guid = C._param.get('guid', str(uuid.uuid4()))
        C.request_id = C._param.get('requestid', C.guid)
        C.quote_mode = C._param.get('quote_mode', 'history') #'realtime' 'history' 'all'
        C.trade_mode = C._param.get('trade_mode', 'backtest') #'simulation' 'trading' 'backtest'
        C.title = C._param.get('title', '')
        if not C.title:
            C.title = os.path.basename(os.path.abspath(C.user_script).replace('.py', ''))

        C.stock_code = C._param.get('stock_code', '')
        C.period = C._param.get('period', '')
        C.start_time = C._param.get('start_time', '')
        C.end_time = C._param.get('end_time', '')
        C.start_time_str = ''
        C.end_time_str = ''
        if type(C.period) == int:
            C.period = {
                0               :'tick'
                , 60000         :'1m'
                , 180000        :'3m'
                , 300000        :'5m'
                , 600000        :'10m'
                , 900000        :'15m'
                , 1800000       :'30m'
                , 3600000       :'1h'
                , 86400000      :'1d'
                , 604800000     :'1w'
                , 2592000000    :'1mon'
                , 7776000000    :'1q'
                , 15552000000   :'1hy'
                , 31536000000   :'1y'
            }.get(C.period, '')
        C.dividend_type = C._param.get('dividend_type', 'none')

        functions._request_id = C.request_id
        xtdata_config.client_guid = C._param.get('clientguid')

        if C.start_time:
            C.start_time_str = C.start_time.replace('-', '').replace(' ', '').replace(':', '')
            C.start_time_num = int(functions.datetime_to_timetag(C.start_time_str))
        if C.end_time:
            C.end_time_str = C.end_time.replace('-', '').replace(' ', '').replace(':', '')
            C.end_time_num = int(functions.datetime_to_timetag(C.end_time_str))

        if 1: #register
            this.create_formula()

        C.init()

        if 1: #fix param
            if '.' in C.stock_code:
                pos = C.stock_code.rfind('.')
                C.stockcode = C.stock_code[0:pos]
                C.market = C.stock_code[pos + 1:].upper()

            if C.stockcode and C.market:
                C.stock_code = C.stockcode + '.' + C.market
            C.period = C.period.lower()

        if C.stockcode == "" or C.market == "":
            raise Exception("股票代码为空")

        if 1: #create view
            if not C._param.get('requestid'):
                this.create_view(C.title)

        if 1: #post initcomplete
            init_result = {}

            config_ar = ['request_id', 'quote_mode', 'trade_mode']
            init_result['config'] = {ar: C.__getattribute__(ar) for ar in config_ar}

            quote_ar = [
                'stock_code', 'stockcode', 'market', 'period'
                , 'start_time', 'end_time', 'dividend_type'
            ]
            init_result['quote'] = {ar: C.__getattribute__(ar) for ar in quote_ar}

            trade_ar = []
            init_result['trade'] = {ar: C.__getattribute__(ar) for ar in trade_ar}

            backtest_ar = [
                'start_time', 'end_time', 'asset', 'margin_ratio', 'slippage_type', 'slippage'
                , 'max_vol_rate', 'comsisson_type', 'open_tax', 'close_tax'
                , 'min_commission', 'open_commission', 'close_commission'
                , 'close_today_commission', 'benchmark'
            ]
            init_result['backtest'] = {ar: C.__getattribute__(ar) for ar in backtest_ar}
            if C.start_time:
                init_result['backtest']['start_time'] = dt.datetime.fromtimestamp(C.start_time_num / 1000).strftime('%Y-%m-%d %H:%M:%S')
            if C.end_time:
                init_result['backtest']['end_time'] = dt.datetime.fromtimestamp(C.end_time_num / 1000).strftime('%Y-%m-%d %H:%M:%S')

            this.call_formula('initcomplete', init_result)

        if 1:
            this.C.register_callback(0)
        return

    def shutdown(this):
        return

    def start(this):
        C = this.C

        if C.quote_mode in ['history', 'all']:
            this.load_main_history()

        C.after_init()
        this.run_bar()

        if C.quote_mode in ['realtime', 'all']:
            this.load_main_realtime()
        return

    def stop(this):
        if this.main_quote_subid:
            xtdata.unsubscribe_quote(this.main_quote_subid)

        this.C.stop()
        return

    def run(this):
        C = this.C

        if C.quote_mode in ['realtime', 'all']:
            xtdata.run()
        return

    def load_main_history(this):
        C = this.C

        data = xtdata.get_market_data_ex(
            field_list = ['time'], stock_list = [C.stock_code], period = C.period
            , start_time = '', end_time = '', count = -1
            , fill_data = False
        )

        C.timelist = list(data[C.stock_code]['time'])
        return

    def load_main_realtime(this):
        C = this.C

        def on_data(data):
            data = data.get(C.stock_code, [])
            if data:
                tt = data[-1]['time']
                this.on_main_quote(tt)
            return

        this.main_quote_subid = xtdata.subscribe_quote(
            stock_code = C.stock_code, period = C.period
            , start_time = '', end_time = '', count = 0
            , callback = on_data
        )
        return

    def on_main_quote(this, timetag):
        if not this.C.timelist or this.C.timelist[-1] < timetag:
            this.C.timelist.append(timetag)
        this.run_bar()
        return

    def run_bar(this):
        C = this.C

        push_timelist = []
        for i in range(max(C.lastrunbarpos, 0), len(C.timelist)):
            C.barpos = i
            bartime = C.timelist[i]
            push_timelist.append(bartime)

            this.call_formula('runbar', {'timelist': [bartime]})

            if (not C.start_time_num or C.start_time_num <= bartime) and (not C.end_time_num or bartime <= C.end_time_num):
                C.handlebar()
            C.lastrunbarpos = i

        if 1:
            push_result = {}
            push_result['timelist'] = push_timelist
            push_result['outputs'] = C.push_result
            C.push_result = {}
            this.call_formula('index', push_result)
        return

    def create_formula(this, callback = None):
        C = this.C
        client = xtdata.get_client()

        data = {
            'formulaname': '', 'stockcode': C.stock_code, 'period': C.period
            , 'starttime': C.start_time_str, 'endtime': C.end_time_str, 'count': 1
            , 'dividendtype': C.dividend_type, 'create': True, 'pyrunmode': 1
            , 'title': C.title
        }

        client.subscribeFormula(C.request_id, bson.BSON.encode(data), callback)

    def call_formula(this, func, data):
        C = this.C
        client = xtdata.get_client()
        bresult = client.callFormula(C.request_id, func, bson.BSON.encode(data))
        return bson.BSON.decode(bresult)

    def create_view(this, title):
        C = this.C
        client = xtdata.get_client()
        data = {'viewtype': 0,'title':title, 'groupid':-1,'stockcode':C.market + C.stockcode,'period':C.period,'dividendtype':C.dividend_type}
        client.createView(C.request_id, bson.BSON.encode(data))
        return




#coding=utf-8
from . import xtconstant

"""
定义Python的数据结构,给Python策略使用
包含股票+信用
"""

class StockAccount(object):
    """
    定义证券账号类, 用于证券账号的报撤单等
    """
    def __new__(cls, account_id, account_type = 'STOCK'):
        """
        :param account_id: 资金账号
        :return: 若资金账号不为字符串，返回类型错误
        """
        if not isinstance(account_id, str):
            return u"资金账号必须为字符串类型"
        return super(StockAccount, cls).__new__(cls)

    def __init__(self, account_id, account_type = 'STOCK'):
        """
        :param account_id: 资金账号
        """
        account_type = account_type.upper()
        for int_type, str_type in xtconstant.ACCOUNT_TYPE_DICT.items():
            if account_type == str_type:
                self.account_type = int_type
                self.account_id = account_id
                return
        raise Exception("不支持的账号类型：{}！".format(account_type))


class XtAsset(object):
    """
    迅投股票账号资金结构
    """
    def __init__(self, account_id, cash, frozen_cash, market_value, total_asset):
        """
        :param account_id: 资金账号
        :param cash: 可用
        :param frozen_cash: 冻结
        :param market_value: 持仓市值
        :param total_asset: 总资产
        """
        self.account_type = xtconstant.SECURITY_ACCOUNT
        self.account_id = account_id
        self.cash = cash
        self.frozen_cash = frozen_cash
        self.market_value = market_value
        self.total_asset = total_asset


class XtOrder(object):
    """
    迅投股票委托结构
    """
    def __init__(self, account_id, stock_code,
                 order_id, order_sysid, order_time, order_type, order_volume,
                 price_type, price, traded_volume, traded_price,
                 order_status, status_msg, strategy_name, order_remark):
        """
        :param account_id: 资金账号
        :param stock_code: 证券代码, 例如"600000.SH"
        :param order_id: 委托编号
        :param order_sysid: 柜台编号
        :param order_time: 报单时间
        :param order_type: 委托类型, 23:买, 24:卖
        :param order_volume: 委托数量, 股票以'股'为单位, 债券以'张'为单位
        :param price_type: 报价类型, 详见帮助手册
        :param price: 报价价格，如果price_type为指定价, 那price为指定的价格，否则填0
        :param traded_volume: 成交数量, 股票以'股'为单位, 债券以'张'为单位
        :param traded_price: 成交均价
        :param order_status: 委托状态
        :param status_msg: 委托状态描述, 如废单原因
        :param strategy_name: 策略名称
        :param order_remark: 委托备注
        """
        self.account_type = xtconstant.SECURITY_ACCOUNT
        self.account_id = account_id
        self.stock_code = stock_code
        self.order_id = order_id
        self.order_sysid = order_sysid
        self.order_time = order_time
        self.order_type = order_type
        self.order_volume = order_volume
        self.price_type = price_type
        self.price = price
        self.traded_volume = traded_volume
        self.traded_price = traded_price
        self.order_status = order_status
        self.status_msg = status_msg
        self.strategy_name = strategy_name
        self.order_remark = order_remark


class XtTrade(object):
    """
    迅投股票成交结构
    """
    def __init__(self, account_id, stock_code,
                 order_type, traded_id, traded_time, traded_price, traded_volume, traded_amount,
                 order_id, order_sysid, strategy_name, order_remark):
        """
        :param account_id: 资金账号
        :param stock_code: 证券代码, 例如"600000.SH"
        :param order_type: 委托类型
        :param traded_id: 成交编号
        :param traded_time: 成交时间
        :param traded_price: 成交均价
        :param traded_volume: 成交数量, 股票以'股'为单位, 债券以'张'为单位
        :param traded_amount: 成交金额
        :param order_id: 委托编号
        :param order_sysid: 柜台编号
        :param strategy_name: 策略名称
        :param order_remark: 委托备注
        """
        self.account_type = xtconstant.SECURITY_ACCOUNT
        self.account_id = account_id
        self.order_type = order_type
        self.stock_code = stock_code
        self.traded_id = traded_id
        self.traded_time = traded_time
        self.traded_price = traded_price
        self.traded_volume = traded_volume
        self.traded_amount = traded_amount
        self.order_id = order_id
        self.order_sysid = order_sysid
        self.strategy_name = strategy_name
        self.order_remark = order_remark


class XtPosition(object):
    """
    迅投股票持仓结构
    """
    def __init__(self, account_id, stock_code,
                 volume, can_use_volume, open_price, market_value,
                 frozen_volume, on_road_volume, yesterday_volume, avg_price):
        """
        :param account_id: 资金账号
        :param stock_code: 证券代码, 例如"600000.SH"
        :param volume: 持仓数量,股票以'股'为单位, 债券以'张'为单位
        :param can_use_volume: 可用数量, 股票以'股'为单位, 债券以'张'为单位
        :param open_price: 开仓价
        :param market_value: 市值
        :param frozen_volume: 冻结数量
        :param on_road_volume: 在途股份
        :param yesterday_volume: 昨夜拥股
        :param avg_price: 成本价
        """
        self.account_type = xtconstant.SECURITY_ACCOUNT
        self.account_id = account_id
        self.stock_code = stock_code
        self.volume = volume
        self.can_use_volume = can_use_volume
        self.open_price = open_price
        self.market_value = market_value
        self.frozen_volume = frozen_volume
        self.on_road_volume = on_road_volume
        self.yesterday_volume = yesterday_volume
        self.avg_price = avg_price


class XtOrderError(object):
    """
    迅投股票委托失败结构
    """
    def __init__(self, account_id, order_id,
                 error_id=None, error_msg=None,
                 strategy_name=None, order_remark=None):
        """
        :param account_id: 资金账号
        :param order_id: 订单编号
        :param error_id: 报单失败错误码
        :param error_msg: 报单失败具体信息
        :param strategy_name: 策略名称
        :param order_remark: 委托备注
        """
        self.account_type = xtconstant.SECURITY_ACCOUNT
        self.account_id = account_id
        self.order_id = order_id
        self.error_id = error_id
        self.error_msg = error_msg
        self.strategy_name = strategy_name
        self.order_remark = order_remark


class XtCancelError(object):
    """
    迅投股票委托撤单失败结构
    """
    def __init__(self, account_id, order_id, market, order_sysid,
                 error_id=None, error_msg=None):
        """
        :param account_id: 资金账号
        :param order_id: 订单编号
        :param market: 交易市场 0:上海 1:深圳
        :param order_sysid: 柜台委托编号
        :param error_id: 撤单失败错误码
        :param error_msg: 撤单失败具体信息
        """
        self.account_type = xtconstant.SECURITY_ACCOUNT
        self.account_id = account_id
        self.order_id = order_id
        self.market = market
        self.order_sysid = order_sysid
        self.error_id = error_id
        self.error_msg = error_msg


class XtOrderResponse(object):
    """
    迅投异步下单接口对应的委托反馈
    """
    def __init__(self, account_id, order_id, strategy_name, order_remark, error_msg, seq):
        """
        :param account_id: 资金账号
        :param order_id: 订单编号
        :param strategy_name: 策略名称
        :param order_remark: 委托备注
        :param seq: 下单请求序号
        """
        self.account_type = xtconstant.SECURITY_ACCOUNT
        self.account_id = account_id
        self.order_id = order_id
        self.strategy_name = strategy_name
        self.order_remark = order_remark
        self.error_msg = error_msg
        self.seq = seq

class XtCancelOrderResponse(object):
    """
    迅投异步委托撤单请求返回结构
    """
    def __init__(self, account_id, cancel_result, order_id, order_sysid, seq):
        """
        :param account_id: 资金账号
        :param cancel_result: 撤单结果
        :param order_id: 订单编号
        :param order_sysid: 柜台委托编号
        :param seq: 撤单请求序号
        """
        self.account_type = xtconstant.SECURITY_ACCOUNT
        self.account_id = account_id
        self.cancel_result = cancel_result
        self.order_id = order_id
        self.order_sysid = order_sysid
        self.seq = seq


class XtCreditOrder(XtOrder):
    """
    迅投信用委托结构
    """
    def __init__(self, account_id, stock_code,
                 order_id, order_time, order_type, order_volume,
                 price_type, price, traded_volume, traded_price,
                 order_status, status_msg, order_remark, contract_no):
        """
        :param account_id: 资金账号
        :param stock_code: 证券代码, 例如"600000.SH"
        :param order_id: 委托编号
        :param order_time: 报单时间
        :param order_type: 委托类型, 23:买, 24:卖
        :param order_volume: 委托数量, 股票以'股'为单位, 债券以'张'为单位
        :param price_type: 报价类型, 详见帮助手册
        :param price: 报价价格，如果price_type为指定价, 那price为指定的价格，否则填0
        :param traded_volume: 成交数量, 股票以'股'为单位, 债券以'张'为单位
        :param traded_price: 成交均价
        :param order_status: 委托状态
        :param status_msg: 委托状态描述, 如废单原因
        :param order_remark: 委托备注
        :param contract_no: 两融合同编号
        """
        self.account_type = xtconstant.CREDIT_ACCOUNT
        self.account_id = account_id
        self.stock_code = stock_code
        self.order_id = order_id
        self.order_time = order_time
        self.order_type = order_type
        self.order_volume = order_volume
        self.price_type = price_type
        self.price = price
        self.traded_volume = traded_volume
        self.traded_price = traded_price
        self.order_status = order_status
        self.status_msg = status_msg
        self.order_remark = order_remark
        self.contract_no = contract_no


class XtCreditDeal(object):
    """
    迅投信用成交结构
    """
    def __init__(self, account_id, stock_code,
                 traded_id, traded_time, traded_price,
                 traded_volume, order_id, contract_no):
        """
        :param account_id: 资金账号
        :param stock_code: 证券代码, 例如"600000.SH"
        :param traded_id: 成交编号
        :param traded_time: 成交时间
        :param traded_price: 成交均价
        :param traded_volume: 成交数量, 股票以'股'为单位, 债券以'张'为单位
        :param order_id: 委托编号
        :param contract_no: 两融合同编号
        """
        self.account_type = xtconstant.CREDIT_ACCOUNT
        self.account_id = account_id
        self.stock_code = stock_code
        self.traded_id = traded_id
        self.traded_time = traded_time
        self.traded_price = traded_price
        self.traded_volume = traded_volume
        self.order_id = order_id
        self.contract_no = contract_no

class XtAccountStatus(object):
    """
    迅投账号状态结构
    """
    def __init__(self, account_id, account_type, status):
        """
        :param account_id: 资金账号
        :param account_type: 账号状态
        :param status: 账号状态，详细见账号状态定义
        """
        self.account_type = account_type
        self.account_id = account_id
        self.status = status


#coding:utf-8

import datetime as dt
import threading

import pandas as pd

from . import functions

class ContextInfo:
    def __init__(this):
        #base
        this.request_id = ''
        this.quote_mode = '' #'realtime' 'history' 'all'
        this.trade_mode = '' #'simulation' 'trading' 'backtest'
        this.title = ''
        this.user_script = ''

        #quote
        this.stock_code = ''
        this.stockcode = ''
        this.market = ''
        this.period = ''
        this.start_time = ''
        this.end_time = ''
        this.dividend_type = ''
        this.start_time_num = None
        this.end_time_num = None

        #bar frame
        this.timelist = []
        this.barpos = -1
        this.lastrunbarpos = -1
        this.result = {}
        this.push_result = {}

        #backtest
        this.asset = 1000000.0  # 初始资金
        this.margin_ratio = 0.05  # 保证金比例
        this.slippage_type = 2  # 滑点类型
        this.slippage = 0.0  # 滑点值
        this.max_vol_rate = 0.0  # 最大成交比例
        this.comsisson_type = 0  # 手续费类型
        this.open_tax = 0.0  # 买入印花税
        this.close_tax = 0.0  # 卖出印花税
        this.min_commission = 0.0  # 最低佣金
        this.open_commission = 0.0  # 买入佣金
        this.close_commission = 0.0  # 平昨佣金
        this.close_today_commission = 0.0  # 平今佣金
        this.benchmark = '000300.SH' # 业绩基准

        this.capital = None
        this.do_back_test = None

        #reserved
        this.refresh_rate = None
        this.fund_name = None
        this.link_fund_name = None
        this.data_info_level = None
        this.time_tick_size = None
        this.subscribe_once = False
        return

    ### qmt strategy frame ###

    def init(this):
        return

    def after_init(this):
        return

    def handlebar(this):
        return

    def stop(this):
        return

    def account_callback(this, account_info):
        return

    def order_callback(this, order_info):
        return

    def deal_callback(this, deal_info):
        return

    def position_callback(this, position_info):
        return

    def orderError_callback(this, passorder_info, msg):
        return

    ### qmt functions - bar ###

    def is_last_bar(this):
        return this.barpos >= len(this.timelist) - 1

    def is_new_bar(this):
        return this.barpos > this.lastbarpos

    def get_bar_timetag(this, barpos = None):
        try:
            return this.timelist[barpos] if barpos is not None else this.timelist[this.barpos]
        except Exception as e:
            return None

    ### qmt functions - graph ###

    def paint(this, name, value, index = -1, drawstyle = 0, color = '', limit = ''):
        vp = {str(this.get_bar_timetag()): value}

        if name not in this.result:
            this.result[name] = {}
        this.result[name].update(vp)

        if name not in this.push_result:
            this.push_result[name] = {}
        this.push_result[name].update(vp)
        return

    ### qmt functions - quote ###

    def subscribe_quote(this, stock_code = '', period = '', dividend_type = '', result_type = '', callback = None):
        if not stock_code:
            stock_code = this.stock_code
        if not period or period == 'follow':
            period = this.period
        if not dividend_type or dividend_type == 'follow':
            dividend_type = this.dividend_type
        return functions.subscribe_quote(stock_code, period, dividend_type, 0, result_type, callback)

    def subscribe_whole_quote(this, code_list, callback = None):
        return functions.subscribe_whole_quote(code_list, callback)

    def unsubscribe_quote(this, subscribe_id):
        return functions.unsubscribe_quote(subscribe_id)

    def get_market_data(
        this, fields = [], stock_code = [], start_time = '', end_time = ''
        , skip_paused = True, period = '', dividend_type = '', count = -1
    ):
        if not stock_code:
            stock_code = [this.stock_code]
        if not period or period == 'follow':
            period = this.period
        if not dividend_type or dividend_type == 'follow':
            dividend_type = this.dividend_type
        if period != 'tick' and count == -1 and len(fields) == 1:
            if not end_time or end_time == 'follow':
                if this.barpos >= 0:
                    end_time = functions.timetag_to_datetime(this.get_bar_timetag(this.barpos))
                    count = -2
        if period == 'tick' and count == -1 and len(fields) == 1 and start_time == '' and end_time == '':
            count = -2

        return functions.get_market_data(
            fields, stock_code, start_time, end_time
            , skip_paused, period, dividend_type, count
        )

    def get_market_data_ex(
        this, fields = [], stock_code = [], period = ''
        , start_time = '', end_time = '', count = -1
        , dividend_type = '', fill_data = True, subscribe = True
    ):
        if not stock_code:
            stock_code = [this.stock_code]
        if not period or period == 'follow':
            period = this.period
        if not dividend_type or dividend_type == 'follow':
            dividend_type = this.dividend_type

        if not this.subscribe_once and subscribe:
            this.subscribe_once = True
            if period != "tick":
                for stk in stock_code:
                    functions.subscribe_quote(stk, period, dividend_type, -1)
            else:
                for stk in stock_code:
                    this.subscribe_whole_quote(stk)

        return functions.get_market_data_ex(
            fields, stock_code, period
            , start_time, end_time, count
            , dividend_type, fill_data, subscribe
        )

    def get_full_tick(this, stock_code = []):
        if not stock_code:
            stock_code = [this.stock_code]
        return functions.get_full_tick(stock_code)

    def get_divid_factors(this, stock_code = '', date = None):
        if not stock_code:
            stock_code = this.stock_code
        return functions.get_divid_factors(stock_code, date)

    ### qmt functions - finance ###

    def get_financial_data(this, field_list, stock_list, start_date, end_date, report_type = 'announce_time'):
        raise 'not implemented, use get_raw_financial_data instead'
        return

    def get_raw_financial_data(this, field_list, stock_list, start_date, end_date, report_type = 'announce_time'):
        return functions.get_raw_financial_data(field_list, stock_list, start_date, end_date, report_type)

    ### qmt functions - option ###

    def get_option_detail_data(this, optioncode):
        return functions.get_option_detail_data(optioncode)

    def get_option_undl_data(this, undl_code_ref):
        return functions.get_option_undl_data(undl_code_ref)

    def get_option_list(this, undl_code,dedate,opttype = "",isavailavle = False):
        return functions.get_option_list(undl_code, dedate, opttype, isavailavle)

    def get_option_iv(this, opt_code):
        return functions.get_opt_iv(opt_code, this.request_id)

    def bsm_price(this, optType, targetPrice, strikePrice, riskFree, sigma, days, dividend = 0):
        optionType = ""
        if(optType.upper() == "C"):
            optionType = "CALL"
        if(optType.upper() == "P"):
            optionType = "PUT"
        if(type(targetPrice) == list):
            result = []
            for price in targetPrice:
                bsmPrice= functions.calc_bsm_price(optionType,strikePrice,float(price),riskFree,sigma,days,dividend, this.request_id)
                bsmPrice = round(bsmPrice,4)
                result.append(bsmPrice)
            return result
        else:
            bsmPrice = functions.calc_bsm_price(optionType,strikePrice,targetPrice,riskFree,sigma,days,dividend, this.request_id)
            result = round(bsmPrice,4)
            return result

    def bsm_iv(this, optType, targetPrice, strikePrice, optionPrice, riskFree, days, dividend = 0):
        if(optType.upper() == "C"):
            optionType = "CALL"
        if(optType.upper() == "P"):
            optionType = "PUT"
        result = functions.calc_bsm_iv(optionType, strikePrice, targetPrice, optionPrice, riskFree, days, dividend, this.request_id)
        result = round(result,4)
        return result

    ### qmt functions - static ###

    def get_instrument_detail(this, stock_code = '', iscomplete = False):
        if not stock_code:
            stock_code = this.stock_code
        return functions.get_instrument_detail(stock_code, iscomplete)

    get_instrumentdetail = get_instrument_detail # compat

    def get_trading_dates(this, stock_code, start_date, end_date, count, period = '1d'):
        return functions.get_trading_dates(stock_code, start_date, end_date, count, period)

    def get_stock_list_in_sector(this, sector_name):
        return functions.get_stock_list_in_sector(sector_name)

    def passorder(this, opType, orderType, accountid, orderCode, prType, modelprice, volume):
        return functions._passorder_impl(opType, orderType, accountid, orderCode, prType, modelprice, volume, this.get_bar_timetag(), this.request_id)

    def set_auto_trade_callback(this, enable):
        return functions._set_auto_trade_callback_impl(enable, this.request_id)

    def set_account(this, accountid):
        return functions.set_account(accountid, this.request_id)

    def get_his_st_data(this, stock_code):
        return functions.get_his_st_data(stock_code)

    ### private ###

    def trade_callback(this, type, result, error):
        class DetailData(object):
            def __init__(self, _obj):
                if _obj:
                    self.__dict__.update(_obj)

        if type == 'accountcallback':
            this.account_callback(DetailData(result))
        elif type == 'ordercallback':
            this.order_callback(DetailData(result))
        elif type == 'dealcallback':
            this.deal_callback(DetailData(result))
        elif type == 'positioncallback':
            this.position_callback(DetailData(result))
        elif type == 'ordererrorcallback':
            this.orderError_callback(DetailData(result.get('passorderArg')), result.get('strMsg'))

        return

    def register_callback(this, reqid):
        functions.register_external_resp_callback(reqid, this.trade_callback)
        return

    def get_callback_cache(this, type):
        return functions._get_callback_cache_impl(type, this.request_id)

    def get_ipo_info(this, start_time = '', end_time = ''):
        return functions.get_ipo_info(start_time, end_time)


#coding:utf-8

import os, sys
import types
import time

from .functions import *
from .contextinfo import ContextInfo
from .stgframe import StrategyLoader


def run_file(user_script, param = {}):
    pypath = param.get('pythonpath')
    if pypath:
        lib_search = [os.path.abspath(p) for p in pypath.split(';')]
        sys.path = lib_search + [p for p in sys.path if p not in lib_search]

    user_module = compile(open(user_script, 'rb').read(), user_script, 'exec', optimize = 2)
    #print({'user_module': user_module})

    try:
        pywentrance = param.get('pywentrance', '')
        user_variable = compile(open(os.path.join(pywentrance, "..", "user_config.py"), "rb").read(),
                                "user_config.py", 'exec', optimize=2)
        exec(user_variable, globals())
    except Exception as e:
        pass

    exec(user_module, globals())

    _C = ContextInfo()
    _C._param = param
    _C.user_script = user_script

    def try_set_func(C, func_name):
        func = globals().get(func_name)
        if func:
            C.__setattr__(func_name, types.MethodType(func, C))
        return

    try_set_func(_C, 'init')
    try_set_func(_C, 'after_init')
    try_set_func(_C, 'handlebar')
    try_set_func(_C, 'stop')

    try_set_func(_C, 'account_callback')
    try_set_func(_C, 'order_callback')
    try_set_func(_C, 'deal_callback')
    try_set_func(_C, 'position_callback')
    try_set_func(_C, 'orderError_callback')

    loader = StrategyLoader()

    loader.C = _C

    loader.init()
    loader.start()
    loader.run()
    loader.stop()
    loader.shutdown()

    while True:
        time.sleep(2)

    return



#coding:utf-8

import os, sys
import time
import traceback
import json
import datetime as dt

from . import xtbson as bson
from . import xtutil


__all__ = [
    'subscribe_quote'
    , 'subscribe_whole_quote'
    , 'unsubscribe_quote'
    , 'run'
    , 'get_market_data'
    , 'get_local_data'
    , 'get_full_tick'
    , 'get_divid_factors'
    , 'get_l2_quote'
    , 'get_l2_order'
    , 'get_l2_transaction'
    , 'download_history_data'
    , 'get_financial_data'
    , 'download_financial_data'
    , 'get_instrument_detail'
    , 'get_instrument_type'
    , 'get_trading_dates'
    , 'get_sector_list'
    , 'get_stock_list_in_sector'
    , 'download_sector_data'
    , 'add_sector'
    , 'remove_sector'
    , 'get_index_weight'
    , 'download_index_weight'
    , 'get_holidays'
    , 'get_trading_calendar'
    , 'get_trade_times'
    #, 'get_industry'
    #, 'get_etf_info'
    #, 'get_main_contract'
    #, 'download_history_contracts'
    , 'download_cb_data'
    , 'get_cb_info'
    , 'create_sector_folder'
    , 'create_sector'
    , 'remove_stock_from_sector'
    , 'reset_sector'
    , 'get_period_list'
]

def try_except(func):
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except Exception:
            exc_type, exc_instance, exc_traceback = sys.exc_info()
            formatted_traceback = ''.join(traceback.format_tb(exc_traceback))
            message = '\n{0} raise {1}:{2}'.format(
                formatted_traceback,
                exc_type.__name__,
                exc_instance
            )
            # raise exc_type(message)
            # print(message)
            return None

    return wrapper


### config

debug_mode = 0

default_data_dir = '../userdata_mini/datadir'
data_dir = default_data_dir


### connection

__client = None
__client_last_spec = ('', None)


def connect(ip = '', port = None, remember_if_success = True):
    global __client
    global data_dir

    if __client:
        if __client.is_connected():
            return __client

        __client.shutdown()
        __client = None
        data_dir = default_data_dir

    from . import xtconn

    if not ip:
        ip = 'localhost'

    if port:
        server_list = [f'{ip}:{port}']
        __client = xtconn.connect_any(server_list)
    else:
        server_list = xtconn.scan_available_server()

        default_addr = 'localhost:58610'
        if not default_addr in server_list:
            server_list.append(default_addr)

        __client = xtconn.connect_any(server_list)

    if not __client or not __client.is_connected():
        raise Exception("无法连接行情服务！")

    if remember_if_success:
        global __client_last_spec
        __client_last_spec = (ip, port)

    data_dir = __client.get_data_dir()
    if data_dir == "":
        data_dir = os.path.join(__client.get_app_dir(), default_data_dir)

    data_dir = os.path.abspath(data_dir)
    return __client


def reconnect(ip = '', port = None, remember_if_success = True):
    global __client
    global data_dir

    if __client:
        __client.shutdown()
        __client = None
        data_dir = default_data_dir

    return connect(ip, port, remember_if_success)


def get_client():
    global __client

    if not __client or not __client.is_connected():
        global __client_last_spec

        ip, port = __client_last_spec
        __client = connect(ip, port, False)

    return __client



### utils

def create_array(shape, dtype_tuple, capsule, size):
    import numpy as np
    import ctypes

    ctypes.pythonapi.PyCapsule_GetPointer.restype = ctypes.POINTER(ctypes.c_char)
    ctypes.pythonapi.PyCapsule_GetPointer.argtypes = [ctypes.py_object, ctypes.c_char_p]
    buff = ctypes.pythonapi.PyCapsule_GetPointer(capsule, None)
    base_type = size * buff._type_

    for dim in shape[::-1]:
        base_type = dim * base_type
    p_arr_type = ctypes.POINTER(base_type)
    obj = ctypes.cast(buff, p_arr_type).contents
    obj._base = capsule
    return np.ndarray(shape = shape, dtype = np.dtype(dtype_tuple), buffer = obj)

from .IPythonApiClient import register_create_nparray
register_create_nparray(create_array)


def __bsoncall_common(interface, func, param):
    return bson.BSON.decode(interface(func, bson.BSON.encode(param)))


### function

def get_industry(industry_name):
    '''
    获取行业成份股，支持申万行业和证监会行业
    :param industry_name: (str)行业名称
    :return: list
    '''
    client = get_client()
    return client.get_industry(industry_name)


def get_stock_list_in_sector(sector_name):
    '''
    获取板块成份股，支持客户端左侧板块列表中任意的板块，包括自定义板块
    :param sector_name: (str)板块名称
    :return: list
    '''
    client = get_client()
    return client.get_stock_list_in_sector(sector_name, 0)


def get_index_weight(index_code):
    '''
    获取某只股票在某指数中的绝对权重
    :param index_code: (str)指数名称
    :return: dict
    '''
    client = get_client()
    return client.get_weight_in_index(index_code)


def get_financial_data(stock_list, table_list=[], start_time='', end_time='', report_type='report_time'):
    '''
     获取财务数据
    :param stock_list: (list)合约代码列表
    :param table_list: (list)报表名称列表
    :param start_time: (str)起始时间
    :param end_time: (str)结束时间
    :param report_type: (str) 时段筛选方式 'announce_time' / 'report_time'
    :return:
        field: list[str]
        date: list[int]
        stock: list[str]
        value: list[list[float]]
    '''
    client = get_client()
    all_table = {
        'Balance' : 'ASHAREBALANCESHEET'
        , 'Income' : 'ASHAREINCOME'
        , 'CashFlow' : 'ASHARECASHFLOW'
        , 'Capital' : 'CAPITALSTRUCTURE'
        , 'HolderNum' : 'SHAREHOLDER'
        , 'Top10Holder' : 'TOP10HOLDER'
        , 'Top10FlowHolder' : 'TOP10FLOWHOLDER'
        , 'PershareIndex' : 'PERSHAREINDEX'
    }

    if not table_list:
        table_list = list(all_table.keys())

    all_table_upper = {table.upper() : all_table[table] for table in all_table}
    req_list = []
    names = {}
    for table in table_list:
        req_table = all_table_upper.get(table.upper(), table)
        req_list.append(req_table)
        names[req_table] = table

    data = {}
    sl_len = 20
    stock_list2 = [stock_list[i : i + sl_len] for i in range(0, len(stock_list), sl_len)]
    for sl in stock_list2:
        data2 = client.get_financial_data(sl, req_list, start_time, end_time, report_type)
        for s in data2:
            data[s] = data2[s]

    import time
    import math
    def conv_date(data, key, key2):
        if key in data:
            tmp_data = data[key]
            if math.isnan(tmp_data):
                if key2 not in data or math.isnan(data[key2]):
                    data[key] = ''
                else:
                    tmp_data = data[key2]
            data[key] = time.strftime('%Y%m%d', time.localtime(tmp_data / 1000))
        return

    result = {}
    import pandas as pd
    for stock in data:
        stock_data = data[stock]
        result[stock] = {}
        for table in stock_data:
            table_data = stock_data[table]
            for row_data in table_data:
                conv_date(row_data, 'm_anntime', 'm_timetag')
                conv_date(row_data, 'm_timetag', '')
                conv_date(row_data, 'declareDate', '')
                conv_date(row_data, 'endDate', '')
            result[stock][names[table]] = pd.DataFrame(table_data)
    return result


def get_market_data_ori(
    field_list = [], stock_list = [], period = '1d'
    , start_time = '', end_time = '', count = -1
    , dividend_type = 'none', fill_data = True, enable_read_from_server = True
):
    client = get_client()
    enable_read_from_local = period in {'1m', '5m', '15m', '30m', '1h', '1d', 'tick'}
    global debug_mode
    return client.get_market_data3(field_list, stock_list, period, start_time, end_time, count, dividend_type, fill_data, 'v2', enable_read_from_local, enable_read_from_server, debug_mode)


def get_market_data(
    field_list = [], stock_list = [], period = '1d'
    , start_time = '', end_time = '', count = -1
    , dividend_type = 'none', fill_data = True
):
    '''
    获取历史行情数据
    :param field_list: 行情数据字段列表，[]为全部字段
        K线可选字段：
            "time"                #时间戳
            "open"                #开盘价
            "high"                #最高价
            "low"                 #最低价
            "close"               #收盘价
            "volume"              #成交量
            "amount"              #成交额
            "settle"              #今结算
            "openInterest"        #持仓量
        分笔可选字段：
            "time"                #时间戳
            "lastPrice"           #最新价
            "open"                #开盘价
            "high"                #最高价
            "low"                 #最低价
            "lastClose"           #前收盘价
            "amount"              #成交总额
            "volume"              #成交总量
            "pvolume"             #原始成交总量
            "stockStatus"         #证券状态
            "openInt"             #持仓量
            "lastSettlementPrice" #前结算
            "askPrice1", "askPrice2", "askPrice3", "askPrice4", "askPrice5" #卖一价~卖五价
            "bidPrice1", "bidPrice2", "bidPrice3", "bidPrice4", "bidPrice5" #买一价~买五价
            "askVol1", "askVol2", "askVol3", "askVol4", "askVol5"           #卖一量~卖五量
            "bidVol1", "bidVol2", "bidVol3", "bidVol4", "bidVol5"           #买一量~买五量
    :param stock_list: 股票代码 "000001.SZ"
    :param period: 周期 分笔"tick" 分钟线"1m"/"5m"/"15m" 日线"1d"
        Level2行情快照"l2quote" Level2行情快照补充"l2quoteaux" Level2逐笔委托"l2order" Level2逐笔成交"l2transaction" Level2大单统计"l2transactioncount" Level2委买委卖队列"l2orderqueue"
        Level1逐笔成交统计一分钟“transactioncount1m” Level1逐笔成交统计日线“transactioncount1d”
        期货仓单“warehousereceipt” 期货席位“futureholderrank” 互动问答“interactiveqa”
    :param start_time: 起始时间 "20200101" "20200101093000"
    :param end_time: 结束时间 "20201231" "20201231150000"
    :param count: 数量 -1全部/n: 从结束时间向前数n个
    :param dividend_type: 除权类型"none" "front" "back" "front_ratio" "back_ratio"
    :param fill_data: 对齐时间戳时是否填充数据，仅对K线有效，分笔周期不对齐时间戳
        为True时，以缺失数据的前一条数据填充
            open、high、low、close 为前一条数据的close
            amount、volume为0
            settle、openInterest 和前一条数据相同
        为False时，缺失数据所有字段填NaN
    :return: 数据集，分笔数据和K线数据格式不同
        period为'tick'时：{stock1 : value1, stock2 : value2, ...}
            stock1, stock2, ... : 合约代码
            value1, value2, ... : np.ndarray 数据列表，按time增序排列
        period为其他K线周期时：{field1 : value1, field2 : value2, ...}
            field1, field2, ... : 数据字段
            value1, value2, ... : pd.DataFrame 字段对应的数据，各字段维度相同，index为stock_list，columns为time_list
    '''
    if period in {'1m', '5m', '15m', '30m', '1h', '1d'}:
        import pandas as pd
        index, data = get_market_data_ori(field_list, stock_list, period, start_time, end_time, count, dividend_type, fill_data)

        result = {}
        for field in data:
            result[field] = pd.DataFrame(data[field], index = index[0], columns = index[1])
        return result

    if period in {'warehousereceipt', 'futureholderrank', 'interactiveqa'}:
        data = get_bson_data(field_list, stock_list, period, start_time, end_time, count)
        return data

    return get_market_data_ori(field_list, stock_list, period, start_time, end_time, count, dividend_type, fill_data)


def get_market_data_ex_ori(
    field_list = [], stock_list = [], period = '1d'
    , start_time = '', end_time = '', count = -1
    , dividend_type = 'none', fill_data = True, enable_read_from_server = True
):
    client = get_client()
    enable_read_from_local = period in {'1m', '5m', '15m', '30m', '1h', '1d', 'tick'}
    global debug_mode
    return client.get_market_data3(field_list, stock_list, period, start_time, end_time, count, dividend_type, fill_data, 'v3', enable_read_from_local, enable_read_from_server, debug_mode)

def convertTimetag(time, is_upper_bound):
    try:
        time1970 = dt.datetime(1970, 1, 1)

        if len(time) == 8:
            time_obj = dt.datetime.strptime(time, '%Y%m%d')
        elif len(time) == 14:
            time_obj = dt.datetime.strptime(time, '%Y%m%d%H%M%S')
        else:
            raise Exception()

        result = int((time_obj - time1970).total_seconds() * 1000) - 28800000

        if len(time) == 8 and is_upper_bound:
            result = result + 86400000 - 1

        return result

    except Exception as e:
        return None

    return None

def get_bson_data(
    field_list, stock_list, period
    , start_time, end_time, count
):
    client = get_client()
    result = {}

    if start_time == '':
        start_time_num = -0x24BCAEE0800
    else:
        start_time_num = convertTimetag(start_time, False)
    if end_time == '':
        end_time_num = 0x1F3FFFFF830
    else:
        end_time_num = convertTimetag(end_time, True)

    if start_time_num is None or end_time_num is None:
        return result

    period_dict = {
        'warehousereceipt': {'time': 'G', 'warehouse': '1', 'receipt': '2'},
        'futureholderrank': {'time': 'G', 'tradingVolumeRank': '1', 'buyPositionRank': '2', 'sellPositionRank': '3'},
        'interactiveqa': {'time': 'G', 'id': '0', 'questionTime': '1', 'question': '2', 'answerTime': '3', 'answer': '4'}
    }
    field_list = field_list if field_list else list(period_dict[period].keys())

    if period in period_dict:
        field_dict = period_dict[period]
    else:
        return result

    data = {
        'stocklist': stock_list,
        'period': period
    }

    result_bson = client.commonControl('getdatafilepath', bson.BSON.encode(data))
    path_result = bson.BSON.decode(result_bson)
    data_path_dict = path_result.get('result', {})

    import pandas as pd

    for stockcode in data_path_dict:
        try:
            bson_data = open(data_path_dict[stockcode], 'rb').read()
            data_list = xtutil.read_from_bson_buffer(bson_data)

            if len(data_list) == 0:
                raise Exception()

            time_index = []
            cut_data_list = []
            for bson_doc in data_list:
                data_time = bson_doc.get('G', None)
                if data_time == None:
                    raise Exception()
                if start_time_num <= data_time and data_time <= end_time_num:
                    time_index.append(data_time)
                    cut_data_list.append(bson_doc)

            data_list = cut_data_list
        except Exception as e:
            result[stockcode] = pd.DataFrame()
            continue

        if count > 0:
            cut_time_index = []
            cut_data_list = []
            index = max(len(data_list) - count, 0)
            for i in range(index, len(data_list)):
                cut_time_index.append(time_index[i])
                cut_data_list.append(data_list[i])
            time_index = cut_time_index
            data_list = cut_data_list

        stockcode_info = {}
        for key in field_list:
            name = field_dict.get(key, None)
            if name is None:
                continue

            value = []
            for bson_doc in data_list:
                field_value = bson_doc.get(name, None)
                value.append(field_value)

            stockcode_info[key] = value

        datetime_index = [timetag_to_datetime(t, '%Y%m%d') for t in time_index]
        stockcode_info_df = pd.DataFrame(stockcode_info, index = datetime_index)
        result[stockcode] = stockcode_info_df

    return result

def get_market_data_ex(
    field_list = [], stock_list = [], period = '1d'
    , start_time = '', end_time = '', count = -1
    , dividend_type = 'none', fill_data = True
):
    if period in {'1m', '5m', '15m', '30m', '1h', '1d'}:
        return _get_market_data_ex_ori_221207(field_list, stock_list, period, start_time, end_time, count, dividend_type, fill_data)

    if period in {'warehousereceipt', 'futureholderrank', 'interactiveqa'}:
        data = get_bson_data(field_list, stock_list, period, start_time, end_time, count)
        return data

    import pandas as pd
    result = {}

    ifield = 'time'
    query_field_list = field_list if (not field_list) or (ifield in field_list) else [ifield] + field_list
    ori_data = get_market_data_ex_ori(query_field_list, stock_list, period, start_time, end_time, count, dividend_type, fill_data)

    if not ori_data:
        return result

    fl = field_list
    stime_fmt = '%Y%m%d' if period == '1d' else '%Y%m%d%H%M%S'
    if fl:
        fl2 = fl if ifield in fl else [ifield] + fl
        for s in ori_data:
            sdata = pd.DataFrame(ori_data[s], columns = fl2)
            sdata2 = sdata[fl]
            sdata2.index = [timetag_to_datetime(t, stime_fmt) for t in sdata[ifield]]
            result[s] = sdata2
    else:
        for s in ori_data:
            sdata = pd.DataFrame(ori_data[s])
            sdata.index = [timetag_to_datetime(t, stime_fmt) for t in sdata[ifield]]
            result[s] = sdata

    return result


def _get_market_data_ex_ori_221207(
    field_list = [], stock_list = [], period = '1d'
    , start_time = '', end_time = '', count = -1
    , dividend_type = 'none', fill_data = True, enable_read_from_server = True
):
    import numpy as np
    import pandas as pd
    client = get_client()
    enable_read_from_local = period in {'1m', '5m', '15m', '30m', '1h', '1d', 'tick'}
    global debug_mode
    ret = client.get_market_data3(field_list, stock_list, period, start_time, end_time, count, dividend_type, fill_data, 'v4', enable_read_from_local,
                                  enable_read_from_server, debug_mode)
    result = {}
    for stock, index, npdatas in ret:
        data = {field: np.frombuffer(b, fi) for field, fi, b in npdatas}
        result[stock] = pd.DataFrame(data=data, index=index)
    return result

def _get_market_data_ex_221207(
    field_list = [], stock_list = [], period = '1d'
    , start_time = '', end_time = '', count = -1
    , dividend_type = 'none', fill_data = True, enable_read_from_server = True
):
    ifield = 'time'
    query_field_list = field_list if (not field_list) or (ifield in field_list) else [ifield] + field_list

    if period in {'1m', '5m', '15m', '30m', '1h', '1d'}:
        ori_data = _get_market_data_ex_ori_221207(query_field_list, stock_list, period, start_time, end_time, count, dividend_type, fill_data, enable_read_from_server)
    else:
        ori_data = get_market_data_ex_ori(query_field_list, stock_list, period, start_time, end_time, count, dividend_type, fill_data, enable_read_from_server)

    import pandas as pd
    result = {}

    fl = field_list

    if fl:
        fl2 = fl if ifield in fl else [ifield] + fl
        for s in ori_data:
            sdata = pd.DataFrame(ori_data[s], columns = fl2)
            sdata2 = sdata[fl]
            sdata2.index = pd.to_datetime((sdata[ifield] + 28800000) * 1000000)
            result[s] = sdata2
    else:
        for s in ori_data:
            sdata = pd.DataFrame(ori_data[s])
            sdata.index = pd.to_datetime((sdata[ifield] + 28800000) * 1000000)
            result[s] = sdata

    return result


get_market_data3 = _get_market_data_ex_221207

def get_data_dir():
    client = get_client()
    global data_dir
    return data_dir

def get_local_data(field_list=[], stock_list=[], period='1d', start_time='', end_time='', count=-1,
                              dividend_type='none', fill_data=True, data_dir=None):
    if data_dir == None:
        data_dir = get_data_dir()

    if period in {'1m', '5m', '15m', '30m', '1h', '1d'}:
        return _get_market_data_ex_ori_221207(field_list, stock_list, period, start_time, end_time, count,
                                              dividend_type, fill_data, False)

    import pandas as pd
    result = {}

    ifield = 'time'
    query_field_list = field_list if (not field_list) or (ifield in field_list) else [ifield] + field_list
    ori_data = get_market_data_ex_ori(query_field_list, stock_list, period, start_time, end_time, count, dividend_type,
                                      fill_data, False)

    if not ori_data:
        return result

    fl = field_list
    stime_fmt = '%Y%m%d' if period == '1d' else '%Y%m%d%H%M%S'
    if fl:
        fl2 = fl if ifield in fl else [ifield] + fl
        for s in ori_data:
            sdata = pd.DataFrame(ori_data[s], columns = fl2)
            sdata2 = sdata[fl]
            sdata2.index = [timetag_to_datetime(t, stime_fmt) for t in sdata[ifield]]
            result[s] = sdata2
    else:
        for s in ori_data:
            sdata = pd.DataFrame(ori_data[s])
            sdata.index = [timetag_to_datetime(t, stime_fmt) for t in sdata[ifield]]
            result[s] = sdata

    return result


def get_l2_quote(field_list=[], stock_code='', start_time='', end_time='', count=-1):
    '''
    level2实时行情
    '''
    global debug_mode
    client = get_client()
    datas = client.get_market_data3(field_list, [stock_code], 'l2quote', start_time, end_time, count, 'none', False, '', False, True, debug_mode)
    if datas:
        return datas[stock_code]
    return None


def get_l2_order(field_list=[], stock_code='', start_time='', end_time='', count=-1):
    '''
    level2逐笔委托
    '''
    global debug_mode
    client = get_client()
    datas = client.get_market_data3(field_list, [stock_code], 'l2order', start_time, end_time, count, 'none', False, '', False, True, debug_mode)
    if datas:
        return datas[stock_code]
    return None


def get_l2_transaction(field_list=[], stock_code='', start_time='', end_time='', count=-1):
    '''
    level2逐笔成交
    '''
    global debug_mode
    client = get_client()
    datas = client.get_market_data3(field_list, [stock_code], 'l2transaction', start_time, end_time, count, 'none', False, '', False, True, debug_mode)
    if datas:
        return datas[stock_code]
    return None


def get_divid_factors(stock_code, start_time='', end_time=''):
    '''
    获取除权除息日及对应的权息
    :param stock_code: (str)股票代码
    :param date: (str)日期
    :return: pd.DataFrame 数据集
    '''
    client = get_client()
    datas = client.get_divid_factors(stock_code, start_time, end_time)
    import pandas as pd
    datas = pd.DataFrame(datas).T
    return datas


@try_except
def getDividFactors(stock_code, date):
    client = get_client()
    resData = client.get_divid_factors(stock_code, date)
    res = {resData[i]: [resData[i + 1][j] for j in
                        range(0, len(resData[i + 1]), 1)] for i in range(0, len(resData), 2)}
    if isinstance(res, dict):
        for k, v in res.items():
            if isinstance(v, list) and len(v) > 5:
                v[5] = int(v[5])
    return res


def get_main_contract(code_market):
    '''
    获取当前期货主力合约
    :param code_market: (str)股票代码
    :return: str
    '''
    client = get_client()
    return client.get_main_contract(code_market)

def datetime_to_timetag(datetime, format = "%Y%m%d%H%M%S"):
    if len(datetime) == 8:
        format = "%Y%m%d"
    timetag = time.mktime(time.strptime(datetime, format))
    return timetag * 1000

def timetag_to_datetime(timetag, format):
    '''
    将毫秒时间转换成日期时间
    :param timetag: (int)时间戳毫秒数
    :param format: (str)时间格式
    :return: str
    '''
    return timetagToDateTime(timetag, format)


@try_except
def timetagToDateTime(timetag, format):
    import time
    timetag = timetag / 1000
    time_local = time.localtime(timetag)
    return time.strftime(format, time_local)


def get_trading_dates(market, start_time='', end_time='', count=-1):
    '''
    根据市场获取交易日列表
    : param market: 市场代码 e.g. 'SH','SZ','IF','DF','SF','ZF'等
    : param start_time: 起始时间 '20200101'
    : param end_time: 结束时间 '20201231'
    : param count: 数据个数，-1为全部数据
    :return list(long) 毫秒数的时间戳列表
    '''
    client = get_client()
    datas = client.get_trading_dates_by_market(market, start_time, end_time, count)
    return datas


def get_full_tick(code_list):
    '''
    获取盘口tick数据
    :param code_list: (list)stock.market组成的股票代码列表
    :return: dict
    {'stock.market': {dict}}
    '''
    client = get_client()
    resp_json = client.get_full_tick(code_list)
    return json.loads(resp_json)


def subscribe_callback_wrapper(callback):
    import traceback
    def subscribe_callback(datas):
        try:
            if type(datas) == bytes:
                datas = bson.BSON.decode(datas)
            if callback:
                callback(datas)
        except:
            print('subscribe callback error:', callback)
            traceback.print_exc()
    return subscribe_callback


def subscribe_quote(stock_code, period='1d', start_time='', end_time='', count=0, callback=None):
    '''
    订阅股票行情数据
    :param stock_code: 股票代码 e.g. "000001.SZ"
    :param start_time: 开始时间，格式YYYYMMDD/YYYYMMDDhhmmss/YYYYMMDDhhmmss.milli，e.g."20200427" "20200427093000" "20200427093000.000"
        若取某日全量历史数据，时间需要具体到秒，e.g."20200427093000"
    :param end_time: 结束时间 同“开始时间”
    :param count: 数量 -1全部/n: 从结束时间向前数n个
    :param period: 周期 分笔"tick" 分钟线"1m"/"5m" 日线"1d"
    :param callback:
        订阅回调函数onSubscribe(datas)
        :param datas: {stock : [data1, data2, ...]} 数据字典
    :return: int 订阅序号
    '''
    if callback:
        callback = subscribe_callback_wrapper(callback)

    meta = {'stockCode': stock_code, 'period': period}
    region = {'startTime': start_time, 'endTime': end_time, 'count': count}

    client = get_client()
    return client.subscribe_quote(bson.BSON.encode(meta), bson.BSON.encode(region), callback)


def subscribe_l2thousand(stock_code, gear_num = 0, callback = None):
    '''
    订阅千档盘口
    '''
    if callback:
        callback = subscribe_callback_wrapper(callback)

    meta = {'stockCode': stock_code, 'period': 'l2thousand'}
    region = {'thousandGearNum': gear_num, 'thousandDetailGear': 0, 'thousandDetailNum': 0}

    client = get_client()
    return client.subscribe_quote(bson.BSON.encode(meta), bson.BSON.encode(region), callback)


def subscribe_whole_quote(code_list, callback=None):
    '''
    订阅全推数据
    :param code_list: 市场代码列表 ["SH", "SZ"]
    :param callback:
        订阅回调函数onSubscribe(datas)
        :param datas: {stock1 : data1, stock2 : data2, ...} 数据字典
    :return: int 订阅序号
    '''
    if callback:
        callback = subscribe_callback_wrapper(callback)

    client = get_client()
    return client.subscribe_whole_quote(code_list, callback)


def unsubscribe_quote(seq):
    '''
    :param seq: 订阅接口subscribe_quote返回的订阅号
    :return:
    '''
    client = get_client()
    return client.unsubscribe_quote(seq)


def run():
    '''阻塞线程接收行情回调'''
    import time
    client = get_client()
    while True:
        time.sleep(3)
        if not client.is_connected():
            raise Exception('行情服务连接断开')
            break
    return

def create_sector_folder(parent_node,folder_name,overwrite = True):
    '''
    创建板块目录节点
    :parent_node str: 父节点,''为'我的' （默认目录）
    :sector_name str: 要创建的板块目录名称
    :overwrite bool:是否覆盖 True为跳过，False为在folder_name后增加数字编号，编号为从1开始自增的第一个不重复的值
    '''
    client = get_client()
    data = {}
    data['parent'] = parent_node
    data['foldername'] = folder_name
    data['overwrite'] = overwrite
    result_bson = client.commonControl('createsectorfolder', bson.BSON.encode(data))
    result = bson.BSON.decode(result_bson)
    return result.get('result')

def create_sector(parent_node,sector_name,overwrite = True):
    '''
    创建板块
    :parent_node str: 父节点,''为'我的' （默认目录）
    :sector_name str: 要创建的板块名
    :overwrite bool:是否覆盖 True为跳过，False为在sector_name后增加数字编号，编号为从1开始自增的第一个不重复的值
    '''
    client = get_client()
    data = {}
    data['parent'] = parent_node
    data['sectorname'] = sector_name
    data['overwrite'] = overwrite
    result_bson = client.commonControl('createsector', bson.BSON.encode(data))
    result = bson.BSON.decode(result_bson)
    return result.get('result')

def get_sector_list():
    '''
    获取板块列表
    :return: (list[str])
    '''
    client = get_client()
    return client.get_sector_list()


def add_sector(sector_name, stock_list):
    '''
    增加自定义板块
    :param sector_name: 板块名称 e.g. "我的自选"
    :param stock_list: (list)stock.market组成的股票代码列表
    '''
    client = get_client()
    data = {}
    data['sectorname'] = sector_name
    data['stocklist'] = stock_list
    result_bson = client.commonControl('addsector', bson.BSON.encode(data))
    result = bson.BSON.decode(result_bson)
    return result.get('result')

def remove_stock_from_sector(sector_name, stock_list):
    '''
    移除板块成分股
    :param sector_name: 板块名称 e.g. "我的自选"
    :stock_list: (list)stock.market组成的股票代码列表
    '''
    client = get_client()
    data = {}
    data['sectorname'] = sector_name
    data['stocklist'] = stock_list
    result_bson = client.commonControl('removestockfromsector', bson.BSON.encode(data))
    result = bson.BSON.decode(result_bson)
    return result.get('result')

def remove_sector(sector_name):
    '''
    删除自定义板块
    :param sector_name: 板块名称 e.g. "我的自选"
    '''
    client = get_client()
    data = {}
    data['sectorname'] = sector_name
    result_bson = client.commonControl('removesector', bson.BSON.encode(data))
    result = bson.BSON.decode(result_bson)
    return result.get('result')

def reset_sector(sector_name, stock_list):
    '''
    重置板块
    :param sector_name: 板块名称 e.g. "我的自选"
    :stock_list: (list)stock.market组成的股票代码列表
    '''
    client = get_client()
    data = {}
    data['sectorname'] = sector_name
    data['stocklist'] = stock_list
    result_bson = client.commonControl('resetsector', bson.BSON.encode(data))
    result = bson.BSON.decode(result_bson)
    return result.get('result')

def get_instrument_detail(stock_code, iscomplete = False):
    '''
    获取合约信息
    :param stock_code: 股票代码 e.g. "600000.SH"
    :return: dict
        ExchangeID(str):合约市场代码
        , InstrumentID(str):合约代码
        , InstrumentName(str):合约名称
        , ProductID(str):合约的品种ID(期货)
        , ProductName(str):合约的品种名称(期货)
        , ExchangeCode(str):交易所代码
        , UniCode(str):统一规则代码
        , CreateDate(int):上市日期(期货)
        , OpenDate(int):IPO日期(股票)
        , ExpireDate(int):退市日或者到期日
        , PreClose(double):前收盘价格
        , SettlementPrice(double):前结算价格
        , UpStopPrice(double):当日涨停价
        , DownStopPrice(double):当日跌停价
        , FloatVolume(double):流通股本
        , TotalVolume(double):总股本
        , LongMarginRatio(double):多头保证金率
        , ShortMarginRatio(double):空头保证金率
        , PriceTick(double):最小变价单位
        , VolumeMultiple(int):合约乘数(对期货以外的品种，默认是1)
        , MainContract(int):主力合约标记
        , LastVolume(int):昨日持仓量
        , InstrumentStatus(int):合约停牌状态
        , IsTrading(bool):合约是否可交易
        , IsRecent(bool):是否是近月合约,
    '''
    client = get_client()
    inst = client.get_instrument_detail(stock_code)
    if not inst:
        return None

    inst = xtutil.read_from_bson_buffer(inst)

    if len(inst) == 1:
        inst = inst[0]
    else:
        return None

    if iscomplete:
        def convNum2Str(field):
            if field in inst and isinstance(inst[field], int):
                inst[field] = str(inst[field])

        convNum2Str('CreateDate')
        convNum2Str('OpenDate')
        convNum2Str('ExpireDate')

        if 'ExtendInfo' in inst:
            def convNum2Str(field):
                if field in inst['ExtendInfo'] and isinstance(inst['ExtendInfo'][field], int):
                    inst['ExtendInfo'][field] = str(inst['ExtendInfo'][field])

            convNum2Str('EndDelivDate')
        return inst

    field_list = [
            'ExchangeID'
            , 'InstrumentID'
            , 'InstrumentName'
            , 'ProductID'
            , 'ProductName'
            , 'ExchangeCode'
            , 'UniCode'
            , 'CreateDate'
            , 'OpenDate'
            , 'ExpireDate'
            , 'PreClose'
            , 'SettlementPrice'
            , 'UpStopPrice'
            , 'DownStopPrice'
            , 'FloatVolume'
            , 'TotalVolume'
            , 'LongMarginRatio'
            , 'ShortMarginRatio'
            , 'PriceTick'
            , 'VolumeMultiple'
            , 'MainContract'
            , 'LastVolume'
            , 'InstrumentStatus'
            , 'IsTrading'
            , 'IsRecent'
        ]
    ret = {}
    for field in field_list:
        ret[field] = inst.get(field)

    exfield_list = [
            'ProductTradeQuota'
            , 'ContractTradeQuota'
            , 'ProductOpenInterestQuota'
            , 'ContractOpenInterestQuota'
        ]
    inst_ex = inst.get('ExtendInfo', {})
    for field in exfield_list:
        ret[field] = inst_ex.get(field)

    def convNum2Str(field):
        if field in ret and isinstance(ret[field], int):
            ret[field] = str(ret[field])
    convNum2Str('CreateDate')
    convNum2Str('OpenDate')

    if ret.get('FloatVolume', None) is None:
        ret['FloatVolume'] = inst.get('FloatVolumn')

    if ret.get('TotalVolume', None) is None:
        ret['TotalVolume'] = inst.get('TotalVolumn')

    return ret


def download_index_weight():
    '''
    下载指数权重数据
    '''
    client = get_client()
    client.down_index_weight()


def download_history_contracts():
    '''
    下载过期合约数据
    '''
    client = get_client()
    client.down_history_contracts()


class TimeListBuilder:
    def __init__(self):
        # param
        self.period = 3600000
        self.open_list = None  # [['093000', '113000'], ['130000', '150000']]

        # build up
        self.cur_date = 0
        self.date_offset = 3600000 * 8
        self.day_time_list = []
        self.cur_index = 0

    def init(self):
        if not self.open_list: return False
        if self.period <= 0: return False

        for scope in self.open_list:
            hour, minute, second = self.parse_time(scope[0])
            start = (((hour * 60) + minute) * 60 + second) * 1000
            hour, minute, second = self.parse_time(scope[1])
            end = (((hour * 60) + minute) * 60 + second) * 1000
            t = start + self.period
            while t <= end:
                self.day_time_list.append(t)
                t += self.period
        self.cur_index = 0

        if not self.day_time_list: return False
        return True

    def parse_time(self, ft):
        ft = int(ft)
        second = ft % 100
        ft = int((ft - second) / 100)
        minute = ft % 100
        ft = int((ft - minute) / 100)
        hour = ft % 100
        return hour, minute, second

    def get(self):
        if self.day_time_list:
            return self.cur_date + self.day_time_list[self.cur_index]
        else:
            return self.cur_date

    def next(self):
        self.cur_index += 1
        if self.cur_index >= len(self.day_time_list):
            self.cur_date += 86400000
            self.cur_index = 0

    def locate(self, t):
        day_time = t % 86400000
        self.cur_date = t - day_time - self.date_offset
        self.cur_index = 0
        for i in range(len(self.day_time_list)):
            te = self.day_time_list[i]
            if t < te:
                self.cur_index = i
                break


class MergeData:
    def __init__(self):
        # param
        self.period = 3600000
        self.open_list = None  # [['093000', '113000'], ['130000', '150000']]
        self.merge_func = None

        # build up
        self.timer = None

        # result
        self.time_list = []
        self.data_list = []

    def init(self):
        self.timer = TimeListBuilder()
        self.timer.open_list = self.open_list
        self.timer.period = self.period
        self.timer.init()

    def push(self, t, data):
        if self.time_list:
            te = self.time_list[-1]
            if t <= te:
                self.data_list[-1] = self.merge_func(self.data_list[-1], data)
            else:
                self.timer.next()
                te = self.timer.get()
                self.time_list.append(te)
                self.data_list.append(data)
        else:
            self.timer.locate(t)

            te = self.timer.get()
            self.time_list.append(te)
            self.data_list.append(data)


def merge_data_sum(data1, data2):
    return data1 + data2


def merge_data_max(data1, data2):
    return max(data1, data2)


def merge_data_min(data1, data2):
    return min(data1, data2)


def merge_data_first(data1, data2):
    return data1


def merge_data_last(data1, data2):
    return data2


def merge_data(time_list, data_list, period, open_list, field):
    merge_func = {}
    merge_func['open'] = merge_data_first
    merge_func['high'] = merge_data_max
    merge_func['low'] = merge_data_min
    merge_func['close'] = merge_data_last
    merge_func['volume'] = merge_data_sum
    merge_func['amount'] = merge_data_sum

    md = MergeData()
    md.period = period
    md.open_list = open_list
    md.merge_func = merge_func[field.lower()]
    md.init()

    for i in range(len(time_list)):
        md.push(time_list[i], data_list[i])

    return md.time_list, md.data_list


def download_history_data(stock_code, period, start_time='', end_time=''):
    '''
    :param stock_code: 股票代码 e.g. "000001.SZ"
    :param period: 周期 分笔"tick" 分钟线"1m"/"5m" 日线"1d"
        Level1逐笔成交统计一分钟“transactioncount1m” Level1逐笔成交统计日线“transactioncount1d”
        期货仓单“warehousereceipt” 期货席位“futureholderrank” 互动问答“interactiveqa”
    :param start_time: 开始时间，格式YYYYMMDD/YYYYMMDDhhmmss/YYYYMMDDhhmmss.milli，e.g."20200427" "20200427093000" "20200427093000.000"
        若取某日全量历史数据，时间需要具体到秒，e.g."20200427093000"
    :param end_time: 结束时间 同上，若是未来某时刻会被视作当前时间
    :return: bool 是否成功
    '''
    client = get_client()
    client.supply_history_data(stock_code, period, start_time, end_time)


supply_history_data = download_history_data


def download_history_data2(stock_list, period, start_time='', end_time='', callback=None):
    '''
    :param stock_code: 股票代码 e.g. "000001.SZ"
    :param period: 周期 分笔"tick" 分钟线"1m"/"5m" 日线"1d"
    :param start_time: 开始时间，格式YYYYMMDD/YYYYMMDDhhmmss/YYYYMMDDhhmmss.milli，e.g."20200427" "20200427093000" "20200427093000.000"
        若取某日全量历史数据，时间需要具体到秒，e.g."20200427093000"
    :param end_time: 结束时间 同上，若是未来某时刻会被视作当前时间
    :return: bool 是否成功
    '''
    client = get_client()

    status = [False, 0, 1, '']
    def on_progress(data):
        try:
            finished = data['finished']
            total = data['total']
            done = (finished >= total)
            status[0] = done
            status[1] = finished
            status[2] = total

            try:
                callback(data)
            except:
                pass

            return done
        except:
            status[0] = True
            status[3] = 'exception'
            return True

    client.supply_history_data2(stock_list, period, start_time, end_time, on_progress)

    import time
    try:
        while not status[0] and client.is_connected():
            time.sleep(0.1)
    except:
        if status[1] < status[2]:
            client.stop_supply_history_data2()
        traceback.print_exc()
    if not client.is_connected():
        raise Exception('行情服务连接断开')
    if status[3]:
        raise Exception('下载数据失败：' + status[3])
    return


def download_financial_data(stock_list, table_list=[], start_time='', end_time=''):
    '''
    :param stock_list: 股票代码列表
    :param table_list: 财务数据表名列表，[]为全部表
        可选范围：['Balance','Income','CashFlow','Capital','Top10FlowHolder','Top10Holder','HolderNum','PershareIndex', 'PerShare']
    :param start_time: 开始时间，格式YYYYMMDD，e.g."20200427"
    :param end_time: 结束时间 同上，若是未来某时刻会被视作当前时间
    '''
    client = get_client()
    if not table_list:
        table_list = ['Balance','Income','CashFlow','Capital','Top10FlowHolder','Top10Holder','HolderNum','PershareIndex', 'PerShare']

    for stock_code in stock_list:
        for table in table_list:
            client.supply_history_data(stock_code, table, start_time, end_time)


def download_financial_data2(stock_list, table_list=[], start_time='', end_time='', callback=None):
    '''
    :param stock_list: 股票代码列表
    :param table_list: 财务数据表名列表，[]为全部表
        可选范围：['Balance','Income','CashFlow','Capital','Top10FlowHolder','Top10Holder','HolderNum','PershareIndex', 'PerShare']
    :param start_time: 开始时间，格式YYYYMMDD，e.g."20200427"
    :param end_time: 结束时间 同上，若是未来某时刻会被视作当前时间
    '''
    client = get_client()
    if not table_list:
        table_list = ['Balance','Income','CashFlow','Capital','Top10FlowHolder','Top10Holder','HolderNum','PershareIndex', 'PerShare']

    data = {}
    data['total'] = len(table_list) * len(stock_list)
    finish = 0
    for stock_code in stock_list:
        for table in table_list:
            client.supply_history_data(stock_code, table, start_time, end_time)

            finish = finish + 1
            try:
                data['finished'] = finish
                callback(data)
            except:
                pass

            if not client.is_connected():
                raise Exception('行情服务连接断开')
                break


def get_instrument_type(stock_code, variety_list = None):
    '''
    判断证券类型
    :param stock_code: 股票代码 e.g. "600000.SH"
    :return: dict{str : bool} {类型名：是否属于该类型}
    '''
    client = get_client()
    v_dct = client.get_stock_type(stock_code)#默认处理得到全部品种的信息
    if not v_dct:
        return {}
    v_dct1 = {}
    if variety_list == None or len(variety_list) == 0:#返回该stock_code所有的品种的T/None(False)
        v_dct1={k: v for k, v in v_dct.items() if v}
        return v_dct1

    for v in variety_list:
        if v in v_dct:
            v_dct1[v] = v_dct[v]
    return v_dct1

get_stock_type = get_instrument_type


def download_sector_data():
    '''
    下载行业板块数据
    '''
    client = get_client()
    client.down_all_sector_data()

def get_holidays():
    '''
    获取节假日列表
    :return: 8位int型日期
    '''
    client = get_client()
    return [str(d) for d in client.get_holidays()]


def get_market_last_trade_date(market):
    client = get_client()
    return client.get_market_last_trade_date(market)

def get_trading_calendar(market, start_time = '', end_time = '', tradetimes = False):
    '''
    获取指定市场交易日历
    :param market: str 市场
    :param start_time: str 起始时间 '20200101'
    :param end_time: str 结束时间 '20201231'
    :param tradetimes: bool 是否包含日内交易时段
    :return:
    '''
    holidays_list = get_holidays()   # 19900101格式的数字
    import datetime
    now = datetime.datetime.combine(datetime.date.today(), datetime.time())
    last = datetime.datetime(now.year + 1, 1, 1)

    client = get_client()
    trading_list = [timetag_to_datetime(x, "%Y%m%d") for x in client.get_trading_dates_by_market(market, start_time, end_time, -1)]

    if start_time == '' and trading_list:
        start_time = trading_list[0]
    start = datetime.datetime.strptime(start_time, "%Y%m%d")

    if end_time == '':
        end_time = now.strftime("%Y%m%d")
    end = min(datetime.datetime.strptime(end_time, "%Y%m%d"), last)

       # 时间戳毫秒
    if not trading_list:
        return []

    if not tradetimes:
        ret_list = trading_list
        while now < end:
            now += datetime.timedelta(days=1)
            if datetime.datetime.isoweekday(now) not in [6, 7]:
                ft = now.strftime("%Y%m%d")
                if ft not in holidays_list:
                    ret_list.append(ft)
        return ret_list
    else:
        ret_map = {}
        trading_times = get_trade_times(market)
        new_trading_times_prev = []  #21-24
        new_trading_times_mid = []  #0-3
        new_trading_times_next = [] #9-15

        for tt in trading_times:
            t0 = tt[0]
            t1 = tt[1]
            t2 = tt[2]
            try:
                if t1 <= 0:
                    new_trading_times_prev.append([t0 + 86400, t1 + 86400, t2])
                elif 0 <= t0 and t1 <= 10800:
                    new_trading_times_mid.append(tt)
                elif t0 <= 0 and t1 <= 10800:
                    new_trading_times_prev.append([t0 + 86400, 86400, t2])
                    new_trading_times_mid.append([0, t1, t2])
                else:
                    new_trading_times_next.append(tt)
            except:
                pass

        end = end + datetime.timedelta(days=1)
        import copy
        prev_open_flag = False
        while start < end:
            weekday = datetime.datetime.isoweekday(start)
            ft = start.strftime("%Y%m%d")
            if weekday not in [6, 7]:
                if ft not in holidays_list:
                    ret_map[ft] = []
                    if prev_open_flag:
                        ret_map[ft].extend(new_trading_times_mid)  # 早盘
                    ret_map[ft].extend(new_trading_times_next)
                    if weekday != 5:
                        if (start + datetime.timedelta(days=1)).strftime("%Y%m%d") not in holidays_list:
                            ret_map[ft].extend(new_trading_times_prev)
                    else:
                        if (start + datetime.timedelta(days=3)).strftime("%Y%m%d") not in holidays_list:
                            ret_map[ft].extend(new_trading_times_prev)
                    prev_open_flag = True
                else:
                    prev_open_flag = False
            start += datetime.timedelta(days=1)
        return ret_map

def get_trade_times(stockcode):
    '''
    返回指定市场或者指定股票的交易时段
    :param stockcode:  市场或者代码.市场  例如 'SH' 或者 '600000.SH'
    :return: 返回交易时段列表，第一位是开始时间，第二位结束时间，第三位交易类型   （2 - 开盘竞价， 3 - 连续交易， 8 - 收盘竞价， 9 - 盘后定价）
    '''
    stockcode_split = stockcode.split('.')
    if len(stockcode_split) == 2:
        ins_dl = get_instrument_detail(stockcode)
        product = ins_dl['ProductID']
        stock = stockcode_split[0]
        market = stockcode_split[1]
        default = 0
    else:
        market = stockcode
        product = ""
        stock = ""
        default = 1

    trader_time = {}
    try:
        with open(os.path.join(data_dir, '..', 'config', 'tradetimeconfig2.json'), 'r') as f:
            trader_time = json.loads(f.read())
    except:
        pass

    ret = []
    import re
    for tdm in trader_time:
        if tdm['default'] == default and tdm['market'] == market:
            if tdm['product'] == [] and tdm['type'] == "":
                ret = tdm['tradetime'] #默认为product为空的 默认值
            if tdm['type'] != "" and re.match(tdm['type'], stock):
                ret = tdm['tradetime']
                break
            if product != "" and product in tdm['product']:
                ret = tdm['tradetime']
                break

    import datetime
    def convert(t):
        if t == "240000" or t == "-240000":
            return 0
        if t[0] == '-':
            parc = datetime.datetime.strptime(t, "-%H%M%S")
            t = datetime.timedelta(hours=-parc.hour, minutes=-parc.minute)
        else:
            parc = datetime.datetime.strptime(t, "%H%M%S")
            t = datetime.timedelta(hours=parc.hour, minutes=parc.minute)
        return int(t.total_seconds())
    ret = [[convert(timepair[0]), convert(timepair[1]), int(timepair[2])] for timepair in ret]
    return ret

def is_stock_type(stock, tag):
    client = get_client()
    return client.is_stock_type(stock, tag)

def download_cb_data():
    client = get_client()
    return client.down_cb_data()
    
def get_cb_info(stockcode):
    client = get_client()
    return client.get_cb_info(stockcode)
    
def get_option_detail_data(optioncode):
    client = get_client()
    inst = client.get_instrument_detail(optioncode)
    if not inst:
        return None
    
    ret = {}
    market = inst.get('ExchangeID')
    if market == 'SHO' or market == "SZO" or (
            (market == "CFFEX" or market == "IF") and inst.get('InstrumentID').find('-') >= 0):
        field_list = [
            'ExchangeID'
            , 'InstrumentID'
            , 'ProductID'
            , 'OpenDate'
            , 'CreateDate'
            , 'ExpireDate'
            , 'PreClose'
            , 'SettlementPrice'
            , 'UpStopPrice'
            , 'DownStopPrice'
            , 'LongMarginRatio'
            , 'ShortMarginRatio'
            , 'PriceTick'
            , 'VolumeMultiple'
            , 'MaxMarketOrderVolume'
            , 'MinMarketOrderVolume'
            , 'MaxLimitOrderVolume'
            , 'MinLimitOrderVolume'
        ]
        ret = {}
        for field in field_list:
            ret[field] = inst.get(field)

        exfield_list = [
            'OptUnit'
            , 'MarginUnit'
            , 'OptUndlCode'
            , 'OptUndlMarket'
            , 'OptExercisePrice'
            , 'NeeqExeType'
            , 'OptUndlRiskFreeRate'
            , 'OptUndlHistoryRate'
            , 'EndDelivDate'
        ]
        inst_ex = inst.get('ExtendInfo', {})
        for field in exfield_list:
            ret[field] = inst_ex.get(field)

        def convNum2Str(field):
            if field in ret and isinstance(ret[field], int):
                ret[field] = str(ret[field])

        convNum2Str('ExpireDate')
        convNum2Str('CreateDate')
        convNum2Str('OpenDate')
        convNum2Str('EndDelivDate')

        ret["optType"] = ""

        instrumentName = inst.get("InstrumentName")
        if instrumentName.find('C') > 0 or instrumentName.find('购') > 0:
            ret["optType"] = "CALL"
        elif instrumentName.find('P') > 0 or instrumentName.find('沽') > 0:
            ret["optType"] = "PUT"
    return ret


def get_option_undl_data(undl_code_ref):
    def get_option_undl(opt_code):
        inst = get_option_detail_data(opt_code)
        if inst and 'OptUndlCode' in inst and 'OptUndlMarket' in inst:
            return inst['OptUndlCode'] + '.' + inst['OptUndlMarket']
        return ''

    if undl_code_ref:
        opt_list = []
        if undl_code_ref.endswith('.SH'):
            if undl_code_ref == "000016.SH" or undl_code_ref == "000300.SH" or undl_code_ref == "000852.SH" or undl_code_ref == "000905.SH":
                opt_list = get_stock_list_in_sector('中金所')
            else:
                opt_list = get_stock_list_in_sector('上证期权')
        if undl_code_ref.endswith('.SZ'):
            opt_list = get_stock_list_in_sector('深证期权')
        data = []
        for opt_code in opt_list:
            undl_code = get_option_undl(opt_code)
            if undl_code == undl_code_ref:
                data.append(opt_code)
        return data
    else:
        opt_list = []
        opt_list += get_stock_list_in_sector('上证期权')
        opt_list += get_stock_list_in_sector('深证期权')
        opt_list += get_stock_list_in_sector('中金所')
        result = {}
        for opt_code in opt_list:
            undl_code = get_option_undl(opt_code)
            if undl_code:
                if undl_code in result:
                    result[undl_code].append(opt_code)
                else:
                    result[undl_code] = [opt_code]
        return result


def get_option_list(undl_code, dedate, opttype = "", isavailavle = False):
    result = []

    marketcodeList = undl_code.split('.')
    if (len(marketcodeList) != 2):
        return []
    undlCode = marketcodeList[0]
    undlMarket = marketcodeList[1]
    market = ""
    if (undlMarket == "SH"):
        if undlCode == "000016" or undlCode == "000300" or undlCode == "000852" or undlCode == "000905":
            market = 'IF'
        else:
            market = "SHO"
    elif (undlMarket == "SZ"):
        market = "SZO"
    if (opttype.upper() == "C"):
        opttype = "CALL"
    elif (opttype.upper() == "P"):
        opttype = "PUT"
    optList = []
    if market == 'SHO':
        optList += get_stock_list_in_sector('上证期权')
        optList += get_stock_list_in_sector('过期上证期权')
    elif market == 'SZO':
        optList += get_stock_list_in_sector('深证期权')
        optList += get_stock_list_in_sector('过期深证期权')
    elif market == 'IF':
        optList += get_stock_list_in_sector('中金所')
        optList += get_stock_list_in_sector('过期中金所')
    for opt in optList:
        if (opt.find(market) < 0):
            continue
        inst = get_option_detail_data(opt)
        if not inst:
            continue
        if (opttype.upper() != "" and opttype.upper() != inst["optType"]):
            continue
        if ((len(dedate) == 6 and inst['ExpireDate'].find(dedate) < 0)):
            continue
        if (len(dedate) == 8):  # option is trade,guosen demand
            createDate = inst['CreateDate']
            openDate = inst['OpenDate']
            if (createDate > '0'):
                openDate = min(openDate, createDate)
            if (openDate < '20150101' or openDate > dedate):
                continue
            endDate = inst['ExpireDate']
            if (isavailavle and endDate < dedate):
                continue
        if inst['OptUndlCode'].find(undlCode) >= 0:
            result.append(opt)
    return result


def get_ipo_info(start_time = '', end_time = ''):
    client = get_client()
    data = client.get_ipo_info(start_time, end_time)
    pylist = [
        'securityCode'          #证券代码
        , 'codeName'            #代码简称
        , 'market'              #所属市场
        , 'actIssueQty'         #发行总量  单位：股
        , 'onlineIssueQty'      #网上发行量  单位：股
        , 'onlineSubCode'       #申购代码
        , 'onlineSubMaxQty'     #申购上限  单位：股
        , 'publishPrice'        #发行价格
        , 'startDate'           #申购开始日期
        , 'onlineSubMinQty'     #最小申购数，单位：股
        , 'isProfit'            #是否已盈利 0：上市时尚未盈利 1：上市时已盈利
        , 'industryPe'          #行业市盈率
        , 'beforePE'            #发行前市盈率
        , 'afterPE'             #发行后市盈率
        , 'listedDate'          #上市日期
        , 'declareDate'         #中签号公布日期
        , 'paymentDate'         #中签缴款日
        , 'lwr'                 #中签率
    ]
    result = []
    for datadict in data:
        resdict = {}
        for field in pylist:
            resdict[field] = datadict.get(field)
        result.append(resdict)
    return result


def get_markets():
    return [
        'SH', 'SZ', 'BJ', 'HK', 'HGT', 'SGT',
        'IF', 'SF', 'DF', 'ZF', 'GF', 'INE',
        'SHO', 'SZO',
        'BKZS', 'WP'
    ]

def get_his_st_data(stock_code):
    fileName = os.path.join(get_data_dir(), '..', 'data', 'SH_XXXXXX_2011_86400000.csv')

    try:
        with open(fileName, "r") as f:
            datas = f.readlines()
    except:
        return {}

    status = []
    for data in datas:
        cols = data.split(',')
        if len(cols) >= 4 and cols[0] == stock_code:
            status.append((cols[2], cols[3]))

    if not status:
        return {}

    result = {}
    i = 0
    while i < len(status):
        start = status[i][0]
        flag = status[i][1]

        i += 1

        end = '20380119'
        if i < len(status):
            end = status[i][0]

        realStatus = ''
        if (flag == '1'):
            realStatus = 'ST'
        elif (flag == '2'):
            realStatus = '*ST'
        elif (flag == '3'):
            realStatus = 'PT'
        else:
            continue

        if realStatus not in result:
            result[realStatus] = []
        result[realStatus].append([start, end])

    return result


def set_instrument_detail(market, stock, stock_name, abbreviation, preclose):
    data = [{"ExchangeID": market, "InstrumentID": stock, "InstrumentName": stock_name, "abbreviation": abbreviation, "PreClose": preclose}]
    client = get_client()
    client.custom_data_control("set_instrument_detail", bson.BSON.encode({"data": data}))

def set_instrument_data2(instruments = []):
    for inst in instruments:
        if type(inst) != dict:
            return "error"
    client = get_client()
    client.custom_data_control("set_instrument_detail", bson.BSON.encode({"data": instruments}))

def create_formula(formula_name, stock_code, period, start_time = '', end_time = '', count = -1, dividend_type = 'none', extend_param = {}, callback = None):
    cl = get_client()

    result = bson.BSON.decode(cl.commonControl('createrequestid', bson.BSON.encode({})))
    request_id = result['result']

    data = {
        'formulaname': formula_name, 'stockcode': stock_code, 'period': period
        , 'starttime': start_time, 'endtime': end_time, 'count': count
        , 'dividendtype': dividend_type, 'extendparam': extend_param
        , 'create': True
    }

    if callback:
        callback = subscribe_callback_wrapper(callback)

    cl.subscribeFormula(request_id, bson.BSON.encode(data), callback)
    return request_id

def subscribe_formula(request_id, callback = None):
    cl = get_client()

    if callback:
        callback = subscribe_callback_wrapper(callback)

    cl.subscribeFormula(request_id, bson.BSON.encode({}), callback)
    return

def unsubscribe_formula(request_id):
    cl = get_client()
    cl.subscribeFormula(request_id)
    return

def call_formula(
    formula_name, stock_code, period
    , start_time = '', end_time = '', count = -1
    , dividend_type = 'none', extend_param = {}
):
    cl = get_client()

    result = bson.BSON.decode(cl.commonControl('createrequestid', bson.BSON.encode({})))
    request_id = result['result']

    data = {
        'formulaname': formula_name, 'stockcode': stock_code, 'period': period
        , 'starttime': start_time, 'endtime': end_time, 'count': count
        , 'dividendtype': dividend_type, 'extendparam': extend_param
        , 'create': True
    }

    data = cl.subscribeFormulaSync(request_id, bson.BSON.encode(data))
    return bson.BSON.decode(data)

gmd = get_market_data
gmd2 = get_market_data_ex
gmd3 = get_market_data3
gld = get_local_data
t2d = timetag_to_datetime
gsl = get_stock_list_in_sector


def reset_market_trading_day_list(market, datas):
    cl = get_client()

    result = __bsoncall_common(
        cl.custom_data_control, 'createmarketchange'
        , {
            'market': market
        }
    )
    cid = result['cid']

    result = __bsoncall_common(
        cl.custom_data_control, 'addtradingdaytochange'
        , {
            'cid': cid
            , 'datas': datas
            , 'coverall': True
        }
    )

    result = __bsoncall_common(
        cl.custom_data_control, 'finishmarketchange'
        , {
            'cid': cid
            #, 'abort': False
            , 'notifyupdate': True
        }
    )
    return


def reset_market_stock_list(market, datas):
    cl = get_client()

    result = __bsoncall_common(
        cl.custom_data_control, 'createmarketchange'
        , {
            'market': market
        }
    )
    cid = result['cid']

    result = __bsoncall_common(
        cl.custom_data_control, 'addstocktochange'
        , {
            'cid': cid
            , 'datas': datas
            , 'coverall': True
        }
    )

    result = __bsoncall_common(
        cl.custom_data_control, 'finishmarketchange'
        , {
            'cid': cid
            #, 'abort': False
            , 'notifyupdate': True
        }
    )
    return


def push_custom_data(meta, datas, coverall = False):
    cl = get_client()
    result = __bsoncall_common(
        cl.custom_data_control, 'pushcustomdata'
        , {
            "meta": meta
            , 'datas': datas
            , 'coverall': coverall
        }
    )
    return

def get_period_list():
    client = get_client()
    result = bson.BSON.decode(client.commonControl('getperiodlist', bson.BSON.encode({})))
    request_id = result['result']
    return request_id

